# Copyright (C) 2020, 2024 Famedly
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
from http import HTTPStatus
from typing import Any

from parameterized import parameterized
from synapse.server import HomeServer
from synapse.util.clock import Clock
from twisted.internet.testing import MemoryReactor

from synapse_invite_checker.types import PermissionConfig, PermissionDefaultSetting
from tests.base import FederatingModuleApiTestCase
from tests.test_utils import INSURANCE_DOMAIN_IN_LIST_FOR_LOCAL


class LocalProModeInviteTest(FederatingModuleApiTestCase):
    """
    These PRO server tests are for invites that happen after the room creation process
    has completed
    """

    def prepare(self, reactor: MemoryReactor, clock: Clock, homeserver: HomeServer):
        super().prepare(reactor, clock, homeserver)
        self.user_a = self.register_user("a", "password")
        self.user_b = self.register_user("b", "password")
        self.user_c = self.register_user("c", "password")
        self.user_d = self.register_user("d", "password")

        self.access_token = self.login("a", "password")
        self.login("b", "password")
        self.login("c", "password")
        self.login("d", "password")

    @parameterized.expand(
        [
            (
                "allow_all_public",
                PermissionDefaultSetting.ALLOW_ALL,
                True,
                HTTPStatus.OK,
            ),
            (
                "allow_all_private",
                PermissionDefaultSetting.ALLOW_ALL,
                False,
                HTTPStatus.OK,
            ),
            (
                "block_all_public",
                PermissionDefaultSetting.BLOCK_ALL,
                True,
                HTTPStatus.FORBIDDEN,
            ),
            (
                "block_all_private",
                PermissionDefaultSetting.BLOCK_ALL,
                False,
                HTTPStatus.FORBIDDEN,
            ),
        ]
    )
    def test_global_permissions(
        self,
        _label: str,
        default_setting: PermissionDefaultSetting,
        is_public: bool,
        expected_result: int,
    ) -> None:
        room_id = self.create_local_room(
            self.user_b,
            [],
            is_public=is_public,
        )
        assert room_id is not None, "Room should have been created"

        # Set the perms
        self.set_permissions_for_user(
            self.user_a,
            PermissionConfig(defaultSetting=default_setting),
        )

        # invite the test user to the other users room
        self.helper.invite(
            room_id,
            self.user_b,
            self.user_a,
            expect_code=expected_result,
            tok=self.map_user_id_to_token[self.user_b],
        )

    @parameterized.expand(
        [
            (
                "allow_all_public",
                PermissionDefaultSetting.ALLOW_ALL,
                True,
                HTTPStatus.FORBIDDEN,
            ),
            (
                "allow_all_private",
                PermissionDefaultSetting.ALLOW_ALL,
                False,
                HTTPStatus.FORBIDDEN,
            ),
            (
                "block_all_public",
                PermissionDefaultSetting.BLOCK_ALL,
                True,
                HTTPStatus.OK,
            ),
            (
                "block_all_private",
                PermissionDefaultSetting.BLOCK_ALL,
                False,
                HTTPStatus.OK,
            ),
        ]
    )
    def test_server_exceptions(
        self,
        _label: str,
        default_setting: PermissionDefaultSetting,
        is_public: bool,
        expected_result: int,
    ) -> None:
        room_id = self.create_local_room(
            self.user_b,
            [],
            is_public=is_public,
        )
        assert room_id is not None, "Room should have been created"

        # Set the perms
        self.set_permissions_for_user(
            self.user_a,
            PermissionConfig(
                defaultSetting=default_setting,
                serverExceptions={self.server_name_for_this_server: {}},
            ),
        )

        # invite the test user to the other users room
        self.helper.invite(
            room_id,
            self.user_b,
            self.user_a,
            expect_code=expected_result,
            tok=self.map_user_id_to_token[self.user_b],
        )

    @parameterized.expand(
        [
            (
                "allow_all_public",  # just a label for test output, ignore
                PermissionDefaultSetting.ALLOW_ALL,  # the global default
                True,  # if the room is public
                # if the expected result should succeed, we use the invert to
                # test the opposite case on the same test run
                # (if "b" would succeed, "c" should fail)
                False,
            ),
            (
                "allow_all_private",
                PermissionDefaultSetting.ALLOW_ALL,
                False,
                False,
            ),
            (
                "block_all_public",
                PermissionDefaultSetting.BLOCK_ALL,
                True,
                True,
            ),
            (
                "block_all_private",
                PermissionDefaultSetting.BLOCK_ALL,
                False,
                True,
            ),
        ]
    )
    def test_user_exceptions(
        self,
        _label: str,
        default_setting: PermissionDefaultSetting,
        is_public: bool,
        expected_result: bool,
    ) -> None:
        # Our test user "a" will be invited to two different rooms, one from user "b"
        # and one from user "c". Since we will have a global default permission that
        # differs, and we are testing userExceptions, the results should be opposite of
        # each other.
        user_exceptions: dict = {self.user_b: {}}

        user_b_expectation = HTTPStatus.OK if expected_result else HTTPStatus.FORBIDDEN
        user_c_expectation = HTTPStatus.FORBIDDEN if expected_result else HTTPStatus.OK
        # Set the perms. By setting them before the invite takes place, it should
        # prevent cross-contamination between other test runs
        self.set_permissions_for_user(
            self.user_a,
            PermissionConfig(
                defaultSetting=default_setting,
                userExceptions=user_exceptions,
            ),
        )

        room_b = self.create_local_room(
            self.user_b,
            [],
            is_public=is_public,
        )
        assert room_b is not None, "Room should have been created"

        # invite the test user to the users rooms that has permission
        self.helper.invite(
            room_b,
            self.user_b,
            self.user_a,
            expect_code=user_b_expectation,
            tok=self.map_user_id_to_token[self.user_b],
        )
        room_c = self.create_local_room(
            self.user_c,
            [],
            is_public=is_public,
        )
        assert room_c is not None, "Room should have been created"

        # invite the test user to the user room that doesn't have permissions
        self.helper.invite(
            room_c,
            self.user_c,
            self.user_a,
            expect_code=user_c_expectation,
            tok=self.map_user_id_to_token[self.user_c],
        )

    def test_invite_to_dm(self) -> None:
        """Tests that a dm with a local user can be created, but nobody else invited"""
        room_id = self.create_local_room(self.user_a, [], is_public=False)
        assert room_id, "Room not created"

        # create DM event
        channel = self.make_request(
            "PUT",
            f"/user/{self.user_a}/account_data/m.direct",
            {
                self.user_b: [room_id],
            },
            access_token=self.access_token,
        )
        assert channel.code == HTTPStatus.OK, channel.result

        # Can't invite other users
        self.helper.invite(
            room=room_id,
            src=self.user_a,
            targ=self.user_c,
            tok=self.access_token,
            expect_code=403,
        )
        self.helper.invite(
            room=room_id,
            src=self.user_a,
            targ=self.user_d,
            tok=self.access_token,
            expect_code=403,
        )
        # But can invite the dm user
        self.helper.invite(
            room=room_id,
            src=self.user_a,
            targ=self.user_b,
            tok=self.access_token,
            expect_code=200,
        )

    def test_invite_to_group(self) -> None:
        """Tests that a group with local users works normally"""
        room_id = self.create_local_room(self.user_a, [], is_public=False)
        assert room_id, "Room not created"

        # create DM event
        channel = self.make_request(
            "PUT",
            f"/user/{self.user_a}/account_data/m.direct",
            {
                self.user_b: ["!not:existing.example.com"],
            },
            access_token=self.access_token,
        )
        assert channel.code == HTTPStatus.OK, channel.result

        # Can invite other users
        self.helper.invite(
            room=room_id,
            src=self.user_a,
            targ=self.user_c,
            tok=self.access_token,
            expect_code=HTTPStatus.OK,
        )
        self.helper.invite(
            room=room_id,
            src=self.user_a,
            targ=self.user_b,
            tok=self.access_token,
            expect_code=200,
        )
        self.helper.invite(
            room=room_id,
            src=self.user_a,
            targ=self.user_d,
            tok=self.access_token,
            expect_code=200,
        )

    def test_invite_to_group_without_dm_event(self) -> None:
        """Tests that a group with local users works normally in case the user has no m.direct set"""
        room_id = self.create_local_room(self.user_a, [], is_public=False)
        assert room_id, "Room not created"

        # Can invite other users
        self.helper.invite(
            room=room_id,
            src=self.user_a,
            targ=self.user_c,
            tok=self.access_token,
            expect_code=200,
        )
        self.helper.invite(
            room=room_id,
            src=self.user_a,
            targ=self.user_b,
            tok=self.access_token,
            expect_code=200,
        )
        self.helper.invite(
            room=room_id,
            src=self.user_a,
            targ=self.user_d,
            tok=self.access_token,
            expect_code=200,
        )


class LocalEpaModeInviteTest(FederatingModuleApiTestCase):
    """
    These EPA server tests are for invites that happen after the room creation process
    has completed
    """

    server_name_for_this_server = INSURANCE_DOMAIN_IN_LIST_FOR_LOCAL

    def prepare(self, reactor: MemoryReactor, clock: Clock, homeserver: HomeServer):
        super().prepare(reactor, clock, homeserver)
        self.user_d = self.register_user("d", "password")
        self.user_e = self.register_user("e", "password")
        self.user_f = self.register_user("f", "password")
        self.access_token = self.login("d", "password")
        self.login("e", "password")
        self.login("f", "password")

    def default_config(self) -> dict[str, Any]:
        conf = super().default_config()
        assert "modules" in conf, "modules missing from config dict during construction"

        # There should only be a single item in the 'modules' list, since this tests that module
        assert len(conf["modules"]) == 1, "more than one module found in config"

        conf["modules"][0].setdefault("config", {}).update({"tim-type": "epa"})
        return conf

    @parameterized.expand(
        [
            (
                "allow_all_private",
                PermissionDefaultSetting.ALLOW_ALL,
                HTTPStatus.FORBIDDEN,
            ),
            (
                "block_all_private",
                PermissionDefaultSetting.BLOCK_ALL,
                HTTPStatus.FORBIDDEN,
            ),
        ]
    )
    def test_global_permissions(
        self,
        _label: str,
        default_setting: PermissionDefaultSetting,
        expected_result: int,
    ) -> None:
        room_id = self.create_local_room(
            self.user_e,
            [],
            is_public=False,
        )
        assert room_id is not None, "Room should have been created"

        # Set the perms
        self.set_permissions_for_user(
            self.user_d,
            PermissionConfig(defaultSetting=default_setting),
        )

        # invite the test user to the other users room
        self.helper.invite(
            room_id,
            self.user_e,
            self.user_d,
            expect_code=expected_result,
            tok=self.map_user_id_to_token[self.user_e],
        )

    @parameterized.expand(
        [
            (
                "allow_all_private",
                PermissionDefaultSetting.ALLOW_ALL,
                HTTPStatus.FORBIDDEN,
            ),
            (
                "block_all_private",
                PermissionDefaultSetting.BLOCK_ALL,
                HTTPStatus.FORBIDDEN,
            ),
        ]
    )
    def test_server_exceptions(
        self,
        _label: str,
        default_setting: PermissionDefaultSetting,
        expected_result: int,
    ) -> None:
        room_id = self.create_local_room(
            self.user_e,
            [],
            is_public=False,
        )
        assert room_id is not None, "Room should have been created"

        # Set the perms
        self.set_permissions_for_user(
            self.user_d,
            PermissionConfig(
                defaultSetting=default_setting,
                serverExceptions={self.server_name_for_this_server: {}},
            ),
        )

        # invite the test user to the other users room
        self.helper.invite(
            room_id,
            self.user_e,
            self.user_d,
            expect_code=expected_result,
            tok=self.map_user_id_to_token[self.user_e],
        )

    @parameterized.expand(
        [
            (
                "allow_all_private",  # just a label for test output, ignore
                PermissionDefaultSetting.ALLOW_ALL,  # the global default
                HTTPStatus.FORBIDDEN,  # the expected return code
            ),
            (
                "block_all_private",
                PermissionDefaultSetting.BLOCK_ALL,
                HTTPStatus.FORBIDDEN,
            ),
        ]
    )
    def test_user_exceptions(
        self,
        _label: str,
        default_setting: PermissionDefaultSetting,
        expected_result: int,
    ) -> None:
        # Our test user "d" will be invited to two different rooms, one from user "e"
        # and one from user "f". For an EPA server, this shouldn't matter as any local
        # invites are denied
        user_exceptions: dict = {self.user_e: {}}

        # Set the perms. By setting them before the invite takes place, it should
        # prevent cross-contamination between other test runs
        self.set_permissions_for_user(
            self.user_d,
            PermissionConfig(
                defaultSetting=default_setting,
                userExceptions=user_exceptions,
            ),
        )

        room_e = self.create_local_room(
            self.user_e,
            [],
            is_public=False,
        )
        assert room_e is not None, "Room should have been created"

        # invite the test user to the users rooms that has permission
        self.helper.invite(
            room_e,
            self.user_e,
            self.user_d,
            expect_code=expected_result,
            tok=self.map_user_id_to_token[self.user_e],
        )
        room_f = self.create_local_room(
            self.user_f,
            [],
            is_public=False,
        )
        assert room_f is not None, "Room should have been created"

        # invite the test user to the user room that doesn't have permissions
        self.helper.invite(
            room_f,
            self.user_f,
            self.user_d,
            expect_code=expected_result,
            tok=self.map_user_id_to_token[self.user_f],
        )

    def test_invite_to_dm_post_room_creation(self) -> None:
        """Tests that a private room as a dm will deny inviting any local users"""
        room_id = self.create_local_room(self.user_d, [], is_public=False)
        assert room_id, "Room not created"

        # create DM event
        channel = self.make_request(
            "PUT",
            f"/user/{self.user_d}/account_data/m.direct",
            {
                self.user_e: [room_id],
            },
            access_token=self.access_token,
        )
        assert channel.code == HTTPStatus.OK, channel.result

        # Can't invite other users
        self.helper.invite(
            room=room_id,
            src=self.user_d,
            targ=self.user_f,
            tok=self.access_token,
            expect_code=403,
        )

        self.helper.invite(
            room=room_id,
            src=self.user_d,
            targ=self.user_e,
            tok=self.access_token,
            expect_code=403,
        )

    def test_invite_to_group_post_room_creation(self) -> None:
        """Tests that a private room for a group will deny inviting any local users, with an unrelated m.direct tag"""
        room_id = self.create_local_room(self.user_d, [], is_public=False)
        assert room_id, "Room not created"

        # create DM event
        channel = self.make_request(
            "PUT",
            f"/user/{self.user_d}/account_data/m.direct",
            {
                self.user_e: ["!not:existing.example.com"],
            },
            access_token=self.access_token,
        )
        assert channel.code == HTTPStatus.OK, channel.result

        # Can't invite other users
        self.helper.invite(
            room=room_id,
            src=self.user_d,
            targ=self.user_f,
            tok=self.access_token,
            expect_code=403,
        )
        self.helper.invite(
            room=room_id,
            src=self.user_d,
            targ=self.user_e,
            tok=self.access_token,
            expect_code=403,
        )

    def test_invite_to_group_without_dm_event_post_room_creation(self) -> None:
        """Tests that a group with local users is denied when the user has no m.direct set"""
        room_id = self.create_local_room(self.user_d, [], is_public=False)
        assert room_id, "Room not created"

        # Can't invite other users
        self.helper.invite(
            room=room_id,
            src=self.user_d,
            targ=self.user_f,
            tok=self.access_token,
            expect_code=403,
        )
        self.helper.invite(
            room=room_id,
            src=self.user_d,
            targ=self.user_e,
            tok=self.access_token,
            expect_code=403,
        )


class DisabledDMCheckInviteTest(FederatingModuleApiTestCase):
    """
    This tests to make sure the DM check can be disabled
    """

    def prepare(self, reactor: MemoryReactor, clock: Clock, homeserver: HomeServer):
        super().prepare(reactor, clock, homeserver)
        self.user_a = self.register_user("a", "password")
        self.access_token = self.login("a", "password")
        self.user_b = self.register_user("b", "password")
        self.user_c = self.register_user("c", "password")
        self.user_d = self.register_user("d", "password")

    def default_config(self) -> dict[str, Any]:
        conf = super().default_config()
        assert "modules" in conf, "modules missing from config dict during construction"

        # There should only be a single item in the 'modules' list, since this tests that module
        assert len(conf["modules"]) == 1, "more than one module found in config"

        conf["modules"][0].setdefault("config", {}).update(
            {"block_invites_into_dms": False}
        )
        return conf

    def test_invite_to_dm(self) -> None:
        """Tests that a dm with a local user can be created, and others can be invited"""
        # This just copies the test from LocalProModeInviteTest but adjusts the expect_code to 200
        room_id = self.create_local_room(self.user_a, [], is_public=False)
        assert room_id, "Room not created"

        # create DM event
        channel = self.make_request(
            "PUT",
            f"/user/{self.user_a}/account_data/m.direct",
            {
                self.user_b: [room_id],
            },
            access_token=self.access_token,
        )
        assert channel.code == HTTPStatus.OK, channel.result

        # Other users can be invited
        self.helper.invite(
            room=room_id,
            src=self.user_a,
            targ=self.user_c,
            tok=self.access_token,
            expect_code=HTTPStatus.OK,
        )
        self.helper.invite(
            room=room_id,
            src=self.user_a,
            targ=self.user_d,
            tok=self.access_token,
            expect_code=200,
        )
        self.helper.invite(
            room=room_id,
            src=self.user_a,
            targ=self.user_b,
            tok=self.access_token,
            expect_code=200,
        )



# Copyright (C) 2025 Famedly
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
from http import HTTPStatus
from typing import Any

from parameterized import parameterized
from synapse.api.constants import EventTypes, HistoryVisibility, JoinRules
from synapse.server import HomeServer
from synapse.util.clock import Clock
from twisted.internet.testing import MemoryReactor

from tests.base import FederatingModuleApiTestCase
from tests.test_utils import INSURANCE_DOMAIN_IN_LIST_FOR_LOCAL


class LocalProModeCreateRoomTest(FederatingModuleApiTestCase):
    """
    These PRO server tests are for room creation process, including invite checking for
    local users and special cases that should be allowed or prevented.
    """

    def prepare(self, reactor: MemoryReactor, clock: Clock, homeserver: HomeServer):
        super().prepare(reactor, clock, homeserver)
        self.pro_user_a = self.register_user("a", "password")
        self.access_token_a = self.login("a", "password")
        self.pro_user_b = self.register_user("b", "password")
        self.access_token_b = self.login("b", "password")
        self.pro_user_c = self.register_user("c", "password")
        self.pro_user_d = self.register_user("d", "password")
        self.access_token_d = self.login("d", "password")

    def default_config(self) -> dict[str, Any]:
        conf = super().default_config()
        assert "modules" in conf, "modules missing from config dict during construction"

        # There should only be a single item in the 'modules' list, since this tests that module
        assert len(conf["modules"]) == 1, "more than one module found in config"

        conf["modules"][0].setdefault("config", {}).update(
            {"default_permissions": {"defaultSetting": "allow all"}}
        )
        conf["server_notices"] = {"system_mxid_localpart": "server", "auto_join": True}
        return conf

    # 'label' as first parameter names the test clearly for failures
    @parameterized.expand([("public", True), ("private", False)])
    def test_create_room(self, label: str, is_public: bool) -> None:
        """Tests room creation with a local user can be created"""
        for invitee in [
            self.pro_user_b,
            self.pro_user_c,
            self.pro_user_d,
        ]:
            room_id = self.create_local_room(
                self.pro_user_a,
                [invitee],
                is_public=is_public,
            )
            assert (
                room_id
            ), f"{label} room from {self.pro_user_a} should be created with invite to: {invitee}"
        for invitee in [
            self.pro_user_a,
            self.pro_user_c,
            self.pro_user_d,
        ]:
            room_id = self.create_local_room(
                self.pro_user_b,
                [invitee],
                is_public=is_public,
            )
            assert (
                room_id
            ), f"{label} room from {self.pro_user_b} should be created with invite to: {invitee}"
        for invitee in [
            self.pro_user_b,
            self.pro_user_c,
            self.pro_user_a,
        ]:
            room_id = self.create_local_room(
                self.pro_user_d,
                [invitee],
                is_public=is_public,
            )
            assert (
                room_id
            ), f"{label} room from {self.pro_user_d} should be created with invite to: {invitee}"

    @parameterized.expand([("public", True), ("private", False)])
    def test_create_room_with_two_invites_fails(
        self, label: str, is_public: bool
    ) -> None:
        """
        Tests that a room can NOT be created when more than one additional member is
        invited during creation
        """
        for invitee_list in [
            [self.pro_user_b, self.pro_user_c],
            [self.pro_user_d, self.pro_user_c],
        ]:
            room_id = self.create_local_room(
                self.pro_user_a,
                invitee_list,
                is_public=is_public,
            )
            assert (
                room_id is None
            ), f"{label} room should not be created with invites to: {invitee_list}"

    def test_create_server_notices_room(self) -> None:
        """
        Test that a server notices room works as expected on pro mode servers
        """
        # send_notice() will automatically create a server notices room and then invite
        # the user it is directed towards. The server notices manager has no method to
        # invite a user during creation of the room
        room_id = self.get_success_or_raise(
            self.hs.get_server_notices_manager().send_notice(
                self.pro_user_d, {"body": "Server Notice message", "msgtype": "m.text"}
            )
        )
        # Retrieving the room_id is a sign that the room was created, the user was
        # invited, and the message was sent
        assert room_id, "Server notices room should have been found"

    @parameterized.expand([("private", False), ("public", True)])
    def test_create_room_then_modify_join_rules(
        self, label: str, is_public: bool
    ) -> None:
        """
        Test that a misbehaving client can not accidentally make their room public after
        the room was created
        """
        room_id = self.create_local_room(self.pro_user_a, [], is_public=is_public)
        assert room_id, f"{label} room should be created"
        # This should be ALLOWED for an already public room, it's silly but is idempotent
        self.helper.send_state(
            room_id,
            EventTypes.JoinRules,
            {"join_rule": JoinRules.PUBLIC},
            tok=self.access_token_a,
            expect_code=HTTPStatus.OK if is_public else HTTPStatus.FORBIDDEN,
        )

    @parameterized.expand([("private", False), ("public", True)])
    def test_create_room_then_modify_history_visibility(
        self, label: str, is_public: bool
    ) -> None:
        """
        Test that a misbehaving client can not accidentally make their room visible
        after the room was created
        """
        room_id = self.create_local_room(self.pro_user_a, [], is_public=is_public)
        assert room_id, f"{label} room should be created"
        # This should be FORBIDDEN for any room
        self.helper.send_state(
            room_id,
            EventTypes.RoomHistoryVisibility,
            {"history_visibility": HistoryVisibility.WORLD_READABLE},
            tok=self.access_token_a,
            expect_code=HTTPStatus.FORBIDDEN,
        )


class LocalEpaModeCreateRoomTest(FederatingModuleApiTestCase):
    """
    These EPA server tests are for room creation process, including invite checking for
    local users and special cases that should be allowed or prevented.
    """

    server_name_for_this_server = INSURANCE_DOMAIN_IN_LIST_FOR_LOCAL

    def prepare(self, reactor: MemoryReactor, clock: Clock, homeserver: HomeServer):
        super().prepare(reactor, clock, homeserver)
        self.epa_user_d = self.register_user("d", "password")
        self.access_token = self.login("d", "password")

        self.epa_user_e = self.register_user("e", "password")
        self.epa_user_f = self.register_user("f", "password")

    def default_config(self) -> dict[str, Any]:
        conf = super().default_config()
        assert "modules" in conf, "modules missing from config dict during construction"

        # There should only be a single item in the 'modules' list, since this tests that module
        assert len(conf["modules"]) == 1, "more than one module found in config"

        conf["modules"][0].setdefault("config", {}).update({"tim-type": "epa"})
        conf["modules"][0].setdefault("config", {}).update(
            {"default_permissions": {"defaultSetting": "allow all"}}
        )
        conf["server_notices"] = {"system_mxid_localpart": "server", "auto_join": True}
        return conf

    @parameterized.expand([("public", True), ("private", False)])
    def test_create_room_fails(self, label: str, is_public: bool) -> None:
        """Tests room creation with a local user will be denied"""
        for invitee in [
            self.epa_user_e,
            self.epa_user_f,
        ]:
            room_id = self.create_local_room(
                self.epa_user_d,
                [invitee],
                is_public=is_public,
            )
            assert (
                room_id is None
            ), f"{label} room should not be created with invite to: {invitee}"

    @parameterized.expand([("public", True), ("private", False)])
    def test_create_room_with_two_invites_fails(
        self, label: str, is_public: bool
    ) -> None:
        """
        Tests that a room can NOT be created when more than one additional member is
        invited during creation
        """
        invitee_list = [self.epa_user_e, self.epa_user_f]
        room_id = self.create_local_room(
            self.epa_user_d,
            invitee_list,
            is_public=is_public,
        )
        assert (
            room_id is None
        ), f"{label} room should not be created with invites to: {invitee_list}"

    def test_create_room_with_modified_join_rules(self) -> None:
        """
        Test that a misbehaving insurance client can not accidentally make their room public
        """
        join_rule = {
            "type": "m.room.join_rules",
            "state_key": "",
            "content": {"join_rule": "public"},
        }
        initial_state = {"initial_state": [join_rule]}

        room_id = self.create_local_room(
            self.epa_user_d, [], is_public=False, override_content=initial_state
        )
        # Without the blocking put in place, this fails for private rooms
        assert room_id is None, "Private room should NOT have been created"

    def test_create_room_with_modified_history_visibility(self) -> None:
        """
        Test that a misbehaving insurance client can not accidentally make their room visible
        """
        history_visibility = {
            "type": EventTypes.RoomHistoryVisibility,
            "state_key": "",
            "content": {"history_visibility": HistoryVisibility.WORLD_READABLE},
        }
        initial_state = {"initial_state": [history_visibility]}

        room_id = self.create_local_room(
            self.epa_user_d, [], is_public=False, override_content=initial_state
        )
        # Without the blocking put in place, this fails for private rooms
        assert room_id is None, "Private room should NOT have been created"

    def test_create_room_then_modify_join_rules(self) -> None:
        """
        Test that a misbehaving insurance client can not accidentally make their room
        public after room was created
        """
        room_id = self.create_local_room(self.epa_user_d, [], is_public=False)
        assert room_id, "Private room should be created"
        # This should be FORBIDDEN
        self.helper.send_state(
            room_id,
            EventTypes.JoinRules,
            {"join_rule": JoinRules.PUBLIC},
            tok=self.access_token,
            expect_code=HTTPStatus.FORBIDDEN,
        )

    def test_create_room_then_modify_history_visibility(self) -> None:
        """
        Test that a misbehaving insurance client can not accidentally make their room
        public after room was created
        """
        room_id = self.create_local_room(self.epa_user_d, [], is_public=False)
        assert room_id, "Private room should be created"
        # This should be FORBIDDEN
        self.helper.send_state(
            room_id,
            EventTypes.RoomHistoryVisibility,
            {"history_visibility": HistoryVisibility.WORLD_READABLE},
            tok=self.access_token,
            expect_code=HTTPStatus.FORBIDDEN,
        )

    def test_create_server_notices_room(self) -> None:
        """
        Test that a server notices room ignores epa restriction rules. This is important
        because server notice rooms are created by a "fake" user on the local server and
        inviting another local server user is supposed to be forbidden. The server
        notice user is considered a system admin account and is therefor exempt from
        this restriction
        """
        # send_notice() will automatically create a server notices room and then invite
        # the user it is directed towards. The server notices manager has no method to
        # invite a user during creation of the room
        room_id = self.get_success_or_raise(
            self.hs.get_server_notices_manager().send_notice(
                self.epa_user_d, {"body": "Server Notice message", "msgtype": "m.text"}
            )
        )
        # Retrieving the room_id is a sign that the room was created, the user was
        # invited, and the message was sent
        assert room_id, "Server notices room should have been found"



# Copyright (C) 2025 Famedly
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
from typing import Any

from parameterized import parameterized
from synapse.server import HomeServer
from synapse.util.clock import Clock
from twisted.internet.testing import MemoryReactor

from tests.base import FederatingModuleApiTestCase
from tests.test_utils import (
    DOMAIN2_IN_LIST,
    DOMAIN_IN_LIST,
    INSURANCE_DOMAIN_IN_LIST,
    INSURANCE_DOMAIN_IN_LIST_FOR_LOCAL,
)


class RemoteProModeCreateRoomTest(FederatingModuleApiTestCase):
    """
    These PRO server tests are for room creation process, to demonstrate that rooms can
    be created when inviting during it's processing.
    NOTE: Event though the server is designated as "block all" the outgoing invites are
    allowed as invites are only checked by the receiving user. This means the room will
    be created, but empty of the other user. Public rooms with invites to a remote user
    will still fail as expected

    Also test special cases that should be allowed or prevented, like trying to invite
    2 users during room creation.
    """

    remote_pro_user = f"@mxid:{DOMAIN_IN_LIST}"
    remote_unlisted_user = f"@gematikuri404:{DOMAIN_IN_LIST}"
    remote_org_user = f"@mxidorg:{DOMAIN_IN_LIST}"
    remote_epa_user = f"@alice:{INSURANCE_DOMAIN_IN_LIST}"
    remote_non_fed_list_user = "@rando:fake-website.com"
    # SERVER_NAME_FROM_LIST = "tim.test.gematik.de"

    def prepare(self, reactor: MemoryReactor, clock: Clock, homeserver: HomeServer):
        super().prepare(reactor, clock, homeserver)
        self.pro_user_a = self.register_user("a", "password")
        self.pro_user_b = self.register_user("b", "password")
        self.login("a", "password")
        self.login("b", "password")

    def default_config(self) -> dict[str, Any]:
        conf = super().default_config()
        assert "modules" in conf, "modules missing from config dict during construction"

        # There should only be a single item in the 'modules' list, since this tests that module
        assert len(conf["modules"]) == 1, "more than one module found in config"

        conf["modules"][0].setdefault("config", {}).update({"tim-type": "pro"})
        conf["modules"][0].setdefault("config", {}).update(
            {"default_permissions": {"defaultSetting": "block all"}}
        )
        return conf

    @parameterized.expand([("public", True, False), ("private", False, True)])
    def test_create_room(
        self, label: str, is_public: bool, expect_creation: bool
    ) -> None:
        """
        Tests room creation from a local User to a remote User behaves as expected
        """
        for remote_user in (self.remote_pro_user, self.remote_epa_user):
            room_id = self.create_local_room(
                self.pro_user_a,
                [remote_user],
                is_public=is_public,
            )
            assert (
                room_id is not None
            ) is expect_creation, f"Creating {label} room from {self.pro_user_a} while inviting {remote_user}: {'denied' if expect_creation else 'allowed'}"

    @parameterized.expand([("public", True), ("private", False)])
    def test_any_user_to_non_fed_domain_create_room_fails(
        self, label: str, is_public: bool
    ) -> None:
        """
        Tests room creation fails from any local User to a remote domain not on the fed list
        """
        for local_user in (self.pro_user_a, self.pro_user_b):
            room_id = self.create_local_room(
                local_user,
                [self.remote_non_fed_list_user],
                is_public=is_public,
            )
            assert (
                room_id is None
            ), f"User '{local_user}'s {label} room with remote non-fed-list domain should not be created"

    @parameterized.expand([("public", True), ("private", False)])
    def test_create_room_with_two_invites_fails(
        self, label: str, is_public: bool
    ) -> None:
        """
        Tests that a room can NOT be created when more than one additional member is
        invited during creation
        """
        # First try with no contact permissions in place, the server default is "block all"
        for invitee_list in [
            # Specifically invite the local user first, as that should always
            # have succeeded
            [self.pro_user_b, self.remote_pro_user],
            [self.pro_user_b, self.remote_epa_user],
            [self.pro_user_b, self.remote_non_fed_list_user],
            # Try with the remote user first too
            [self.remote_pro_user, self.pro_user_b],
            [self.remote_epa_user, self.pro_user_b],
            [self.remote_non_fed_list_user, self.pro_user_b],
        ]:
            room_id = self.create_local_room(
                self.pro_user_a,
                invitee_list,
                is_public=is_public,
            )
            assert (
                room_id is None
            ), f"User-HBA {label} room should not be created(before permission) with invites to: {invitee_list}"

        for remote_user_to_add in (
            self.remote_pro_user,
            self.remote_epa_user,
            self.remote_non_fed_list_user,
            self.pro_user_b,
        ):
            # Grant the exception, but it won't matter
            self.add_permission_to_a_user(remote_user_to_add, self.pro_user_a)

        # Then try with contact permissions added
        for invitee_list in [
            [self.pro_user_b, self.remote_pro_user],
            [self.pro_user_b, self.remote_epa_user],
            [self.pro_user_b, self.remote_non_fed_list_user],
            [self.remote_pro_user, self.pro_user_b],
            [self.remote_epa_user, self.pro_user_b],
            [self.remote_non_fed_list_user, self.pro_user_b],
        ]:
            room_id = self.create_local_room(
                self.pro_user_a,
                invitee_list,
                is_public=is_public,
            )
            assert (
                room_id is None
            ), f"User-HBA {label} room should not be created(after permission) with invites to: {invitee_list}"


class RemoteEpaModeCreateRoomTest(FederatingModuleApiTestCase):
    """
    These EPA server tests are for room creation process, including invite checking for
    REMOTE users and special cases that should be allowed or prevented.

    ePA mode servers should only have insured Users.
    Per https://gemspec.gematik.de/docs/gemSpec/gemSpec_TI-M_ePA/latest/#AF_10233 and
    its two additions(A_20704 and A_20704)
    an invitation to a room where both parties are insured should be denied.
    """

    remote_pro_user = f"@mxid:{DOMAIN_IN_LIST}"
    remote_pro_user_2 = f"@gematikuri2org:{DOMAIN2_IN_LIST}"
    remote_epa_user = f"@alice:{INSURANCE_DOMAIN_IN_LIST}"
    remote_non_fed_list_user = "@rando:fake-website.com"
    server_name_for_this_server = INSURANCE_DOMAIN_IN_LIST_FOR_LOCAL

    def prepare(self, reactor: MemoryReactor, clock: Clock, homeserver: HomeServer):
        super().prepare(reactor, clock, homeserver)
        self.epa_user_d = self.register_user("d", "password")
        self.epa_user_e = self.register_user("e", "password")
        self.login("d", "password")
        self.login("e", "password")

    def default_config(self) -> dict[str, Any]:
        conf = super().default_config()
        assert "modules" in conf, "modules missing from config dict during construction"

        # There should only be a single item in the 'modules' list, since this tests that module
        assert len(conf["modules"]) == 1, "more than one module found in config"

        conf["modules"][0].setdefault("config", {}).update({"tim-type": "epa"})
        conf["modules"][0].setdefault("config", {}).update(
            {"default_permissions": {"defaultSetting": "allow all"}}
        )
        return conf

    @parameterized.expand([("public", True), ("private", False)])
    def test_epa_to_pro_create_room(self, label: str, is_public: bool) -> None:
        """
        Tests room creation from a local insured User to a remote Pro-User behaves as expected
        """
        room_id = self.create_local_room(
            self.epa_user_d,
            [self.remote_pro_user],
            is_public=is_public,
        )
        assert (
            (room_id is None) if is_public else room_id
        ), f"Epa-User {label} room with remote Pro-User should be: {'denied' if is_public else 'allowed'}"

    @parameterized.expand([("public", True), ("private", False)])
    def test_epa_to_epa_create_room_fails(self, label: str, is_public: bool) -> None:
        """
        Tests room creation from a local insured User to a remote insured User
        fails as expected.
        """
        room_id = self.create_local_room(
            self.epa_user_d,
            [self.remote_epa_user],
            is_public=is_public,
        )
        assert (
            room_id is None
        ), f"User-ePA {label} room with remote insured should not be created(before permissions)"

        # Touching the permission(since the server is "allow all") should not help them
        # get a room made
        self.add_permission_to_a_user(self.remote_epa_user, self.epa_user_d)

        room_id = self.create_local_room(
            self.epa_user_d,
            [self.remote_epa_user],
            is_public=is_public,
        )
        assert (
            room_id is None
        ), f"User-ePA {label} room with remote insured should not be created(after permissions)"

    @parameterized.expand([("public", True), ("private", False)])
    def test_epa_to_non_fed_domain_create_any_room_fails(
        self, label: str, is_public: bool
    ) -> None:
        """
        Tests room creation from a local insured User to a remote domain not on the fed list fails
        """
        room_id = self.create_local_room(
            self.epa_user_d,
            [self.remote_non_fed_list_user],
            is_public=is_public,
        )
        assert (
            room_id is None
        ), f"User-ePA {label} room with remote non-fed-list domain should not be created(before permissions)"

        # Touching the permission(since the server is "allow all") should not help them
        # get a room made
        self.add_permission_to_a_user(self.remote_non_fed_list_user, self.epa_user_d)

        room_id = self.create_local_room(
            self.epa_user_d,
            [self.remote_non_fed_list_user],
            is_public=is_public,
        )
        assert (
            room_id is None
        ), f"User-ePA {label} room with remote non-fed-list domain should not be created(after permissions)"

    @parameterized.expand([("public", True), ("private", False)])
    def test_create_room_with_two_invites_fails(
        self, label: str, is_public: bool
    ) -> None:
        """
        Tests that room creation fails with more than one included invite
        """
        # User "d" got contaminated in other tests with permissions, use a clean user
        # to create rooms
        for invitee_list in [
            [self.remote_pro_user_2, self.remote_pro_user],
            [self.remote_pro_user_2, self.remote_epa_user],
            [self.remote_pro_user_2, self.remote_non_fed_list_user],
        ]:
            room_id = self.create_local_room(
                self.epa_user_e,
                invitee_list,
                is_public=is_public,
            )
            assert (
                room_id is None
            ), f"User-ePA {label} room should not be created(before permission) with invites to: {invitee_list}"

        # Touching the permission(since the server is "allow all") should not help them
        # get a room made
        for remote_user_to_add in (
            self.remote_pro_user,
            self.remote_epa_user,
            self.remote_non_fed_list_user,
            self.remote_pro_user_2,
        ):
            self.add_permission_to_a_user(remote_user_to_add, self.epa_user_e)

        for invitee_list in [
            [self.remote_pro_user_2, self.remote_pro_user],
            [self.remote_pro_user_2, self.remote_epa_user],
            [self.remote_pro_user_2, self.remote_non_fed_list_user],
        ]:
            room_id = self.create_local_room(
                self.epa_user_e,
                invitee_list,
                is_public=is_public,
            )
            assert (
                room_id is None
            ), f"User-ePA {label} room should not be created(after permission) with invites to: {invitee_list}"



# Copyright (C) 2020,2023 Famedly
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
from dataclasses import dataclass, field

from synapse_invite_checker.types import DefaultPermissionConfig, TimType


@dataclass
class InsuredOnlyRoomScanConfig:
    grace_period_ms: int = 0
    invites_grace_period_ms: int = 0
    enabled: bool = False


@dataclass
class InactiveRoomScanConfig:
    grace_period_ms: int = 0
    enabled: bool = False


@dataclass
class InviteCheckerConfig:
    default_permissions: DefaultPermissionConfig
    title: str = "Invite Checker module by Famedly"
    description: str = "Invite Checker module by Famedly"
    contact: str = "info@famedly.com"
    federation_list_url: str = ""
    federation_list_client_cert: str = ""
    federation_list_require_mtls: bool = True
    gematik_ca_baseurl: str = ""
    tim_type: TimType = TimType.PRO
    allowed_room_versions: list[str] = field(default_factory=list)
    room_scan_run_interval_ms: int = 0
    insured_room_scan_options: InsuredOnlyRoomScanConfig = field(
        default_factory=InsuredOnlyRoomScanConfig
    )
    inactive_room_scan_options: InactiveRoomScanConfig = field(
        default_factory=InactiveRoomScanConfig
    )
    override_public_room_federation: bool = True
    prohibit_world_readable_rooms: bool = True
    block_invites_into_dms: bool = True
    limit_reactions: bool = True



# Copyright 2018 New Vector Ltd
# Copyright 2020 The Matrix.org Foundation C.I.C
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any

import synapse.server
from synapse.api.constants import EventTypes
from synapse.api.room_versions import KNOWN_ROOM_VERSIONS
from synapse.events import EventBase
from synapse.events.snapshot import EventContext

"""
Utility functions for poking events into the storage of the server under test.
"""


async def inject_member_event(
    hs: synapse.server.HomeServer,
    room_id: str,
    sender: str,
    membership: str,
    target: str | None = None,
    extra_content: dict | None = None,
    **kwargs: Any,
) -> EventBase:
    """Inject a membership event into a room."""
    if target is None:
        target = sender

    content = {"membership": membership}
    if extra_content:
        content.update(extra_content)

    return await inject_event(
        hs,
        room_id=room_id,
        type=EventTypes.Member,
        sender=sender,
        state_key=target,
        content=content,
        **kwargs,
    )


async def inject_event(
    hs: synapse.server.HomeServer,
    room_version: str | None = None,
    prev_event_ids: list[str] | None = None,
    **kwargs: Any,
) -> EventBase:
    """Inject a generic event into a room

    Args:
        hs: the homeserver under test
        room_version: the version of the room we're inserting into.
            if not specified, will be looked up
        prev_event_ids: prev_events for the event. If not specified, will be looked up
        kwargs: fields for the event to be created
    """
    event, context = await create_event(hs, room_version, prev_event_ids, **kwargs)

    persistence = hs.get_storage_controllers().persistence
    assert persistence is not None

    await persistence.persist_event(event, context)

    return event


async def create_event(
    hs: synapse.server.HomeServer,
    room_version: str | None = None,
    prev_event_ids: list[str] | None = None,
    **kwargs: Any,
) -> tuple[EventBase, EventContext]:
    if room_version is None:
        room_version = await hs.get_datastores().main.get_room_version_id(
            kwargs["room_id"]
        )

    builder = hs.get_event_builder_factory().for_room_version(
        KNOWN_ROOM_VERSIONS[room_version], kwargs
    )
    (
        event,
        unpersisted_context,
    ) = await hs.get_event_creation_handler().create_new_client_event(
        builder, prev_event_ids=prev_event_ids
    )

    context = await unpersisted_context.persist(event)

    return event, context


async def mark_event_as_partial_state(
    hs: synapse.server.HomeServer,
    event_id: str,
    room_id: str,
) -> None:
    """
    (Falsely) mark an event as having partial state.

    Naughty, but occasionally useful when checking that partial state doesn't
    block something from happening.

    If the event already has partial state, this insert will fail (event_id is unique
    in this table).
    """
    store = hs.get_datastores().main
    await store.db_pool.simple_upsert(
        table="partial_state_rooms",
        keyvalues={"room_id": room_id},
        values={},
        insertion_values={"room_id": room_id},
    )

    await store.db_pool.simple_insert(
        table="partial_state_events",
        values={
            "room_id": room_id,
            "event_id": event_id,
        },
    )



# Copyright (C) 2020, 2024 Famedly
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
import re
from typing import Any

import pytest
from synapse.module_api import NOT_SPAM, errors
from synapse.server import HomeServer
from synapse.types import UserID
from synapse.util.clock import Clock
from twisted.internet import defer
from twisted.internet.testing import MemoryReactor

from synapse_invite_checker.types import (
    GroupName,
    PermissionConfig,
    PermissionDefaultSetting,
)
from tests.base import FederatingModuleApiTestCase
from tests.test_utils import (
    DOMAIN3_IN_LIST,
    DOMAIN_IN_LIST,
    INSURANCE_DOMAIN_IN_LIST,
    INSURANCE_DOMAIN_IN_LIST_FOR_LOCAL,
)


class RemoteProModeInviteTest(FederatingModuleApiTestCase):
    """
    These PRO server tests are for invites that happen after the room creation process
    has completed
    """

    # SERVER_NAME_FROM_LIST = "tim.test.gematik.de"

    def prepare(self, reactor: MemoryReactor, clock: Clock, homeserver: HomeServer):
        super().prepare(reactor, clock, homeserver)
        self.pro_user_a = self.register_user("a", "password")
        self.pro_user_b = self.register_user("b", "password")
        self.pro_user_c = self.register_user("c", "password")
        self.pro_user_d = self.register_user("d", "password")
        self.non_existent_user = f"@notarealuser:{self.server_name_for_this_server}"
        self.login("a", "password")
        self.login("b", "password")
        self.login("c", "password")
        self.login("d", "password")

    def may_invite(self, inviter: str, invitee: str, roomid: str):
        req = defer.ensureDeferred(
            self.hs.get_module_api()._callbacks.spam_checker.user_may_invite(
                inviter, invitee, roomid
            )
        )
        self.wait_on_thread(req)
        ret = self.get_success(req)
        if ret == NOT_SPAM:
            return NOT_SPAM
        return ret[0]  # return first code instead of all of them to make assert easier

    def test_invite_from_remote_user(self) -> None:
        """
        Tests that an invite from a remote user.
        """
        for remote_user_id in [
            f"@example:{DOMAIN_IN_LIST}",
            f"@example:{DOMAIN3_IN_LIST}",
            f"@mxid404:{DOMAIN_IN_LIST}",
        ]:
            for local_user in [
                self.pro_user_a,
                self.pro_user_b,
                self.pro_user_c,
                self.pro_user_d,
            ]:
                # default
                assert (
                    self.may_invite(remote_user_id, local_user, "!madeup:example.com")
                    == NOT_SPAM
                ), f"'{remote_user_id}' should be ALLOWED to invite {local_user}"

                # Explicit allow all
                self.set_permissions_for_user(
                    local_user,
                    PermissionConfig(defaultSetting=PermissionDefaultSetting.ALLOW_ALL),
                )
                assert (
                    self.may_invite(remote_user_id, local_user, "!madeup:example.com")
                    == NOT_SPAM
                ), f"'{remote_user_id}' should be ALLOWED to invite {local_user} (explicit all)"

                # Explicit block all
                self.set_permissions_for_user(
                    local_user,
                    PermissionConfig(defaultSetting=PermissionDefaultSetting.BLOCK_ALL),
                )
                assert (
                    self.may_invite(remote_user_id, local_user, "!madeup:example.com")
                    == errors.Codes.FORBIDDEN
                ), f"'{remote_user_id}' should be BLOCKED from inviting {local_user} (explicit all)"

                # Explicit allow single
                self.set_permissions_for_user(
                    local_user,
                    PermissionConfig(
                        defaultSetting=PermissionDefaultSetting.BLOCK_ALL,
                        userExceptions={remote_user_id: {}},
                    ),
                )
                assert (
                    self.may_invite(remote_user_id, local_user, "!madeup:example.com")
                    == NOT_SPAM
                ), f"'{remote_user_id}' should be ALLOWED to invite {local_user} (explicit single)"

                # Explicit block single
                self.set_permissions_for_user(
                    local_user,
                    PermissionConfig(
                        defaultSetting=PermissionDefaultSetting.ALLOW_ALL,
                        userExceptions={remote_user_id: {}},
                    ),
                )
                assert (
                    self.may_invite(remote_user_id, local_user, "!madeup:example.com")
                    == errors.Codes.FORBIDDEN
                ), f"'{remote_user_id}' should be BLOCKED from inviting {local_user} (explicit single)"

                # Explicit allow server
                self.set_permissions_for_user(
                    local_user,
                    PermissionConfig(
                        defaultSetting=PermissionDefaultSetting.BLOCK_ALL,
                        serverExceptions={
                            UserID.from_string(remote_user_id).domain: {}
                        },
                    ),
                )
                assert (
                    self.may_invite(remote_user_id, local_user, "!madeup:example.com")
                    == NOT_SPAM
                ), f"'{remote_user_id}' should be ALLOWED to invite {local_user} (explicit server)"

                # Explicit block server
                self.set_permissions_for_user(
                    local_user,
                    PermissionConfig(
                        defaultSetting=PermissionDefaultSetting.ALLOW_ALL,
                        serverExceptions={
                            UserID.from_string(remote_user_id).domain: {}
                        },
                    ),
                )
                assert (
                    self.may_invite(remote_user_id, local_user, "!madeup:example.com")
                    == errors.Codes.FORBIDDEN
                ), f"'{remote_user_id}' should be BLOCKED from inviting {local_user} (explicit server)"

    def test_invite_from_remote_outside_of_fed_list(self) -> None:
        """Tests that an invite from a remote server not in the federation list gets denied"""
        for remote_user_id in [
            f"@example:not-{DOMAIN_IN_LIST}",
            f"@example2:not-{DOMAIN_IN_LIST}",
            "@madeupuser:thecornerstore.de",
            "@unknown:not.in.fed",
        ]:
            for local_user in [
                self.pro_user_a,
                self.pro_user_b,
                self.pro_user_c,
                self.pro_user_d,
            ]:
                assert (
                    self.may_invite(remote_user_id, local_user, "!madeup:example.com")
                    == errors.Codes.FORBIDDEN
                ), f"'{remote_user_id}' should be FORBIDDEN to invite {local_user}(before permission)"

            for local_user in [
                self.pro_user_a,
                self.pro_user_b,
                self.pro_user_c,
                self.pro_user_d,
            ]:
                # Add permissions, but it shouldn't matter
                self.add_permission_to_a_user(remote_user_id, local_user)
                assert (
                    self.may_invite(remote_user_id, local_user, "!madeup:example.com")
                    == errors.Codes.FORBIDDEN
                ), f"'{remote_user_id}' should be FORBIDDEN to invite {local_user}(after permission)"

    def test_invite_to_nonexistent_local_user(self) -> None:
        """Tests that an invite to a local user that does not exist gets denied"""
        for remote_user_id in [
            f"@example:{DOMAIN_IN_LIST}",
            f"@example:{DOMAIN3_IN_LIST}",
            f"@mxid404:{DOMAIN_IN_LIST}",
        ]:
            assert (
                self.may_invite(
                    remote_user_id, self.non_existent_user, "!madeup:example.com"
                )
                == errors.Codes.FORBIDDEN
            ), f"'{remote_user_id}' should be FORBIDDEN to invite {self.non_existent_user}(before permission)"

            # Adding permissions will not help, as the user doesn't exist
            with pytest.raises(
                ValueError,
                match=re.escape(
                    "User @notarealuser:tim.test.gematik.de does not exist on this server."
                ),
            ):
                # Raises with this error:
                # `ValueError: User @notarealuser:tim.test.gematik.de does not exist on this server.`
                # Weirdly, it's not the retrieval of the account data that triggers it,
                # but is the trying to put new account data
                self.add_permission_to_a_user(remote_user_id, self.non_existent_user)
            # May as well give it another go anyway, just to make sure getting the
            # non-existent account data didn't cause it to suddenly work
            assert (
                self.may_invite(
                    remote_user_id, self.non_existent_user, "!madeup:example.com"
                )
                == errors.Codes.FORBIDDEN
            ), f"'{remote_user_id}' should be FORBIDDEN to invite {self.non_existent_user}(after permission)"

    def test_remote_invite_from_an_insurance_domain(self) -> None:
        """
        Test that an insured user can invite a publicly listed practitioner or organization
        (but not a user who blocked the insurance group)
        """
        for remote_user_id in (
            f"@unknown:{INSURANCE_DOMAIN_IN_LIST}",
            f"@rando-32-b52:{INSURANCE_DOMAIN_IN_LIST}",
        ):
            assert (
                self.may_invite(remote_user_id, self.pro_user_b, "!madeup:example.com")
                == NOT_SPAM
            ), f"'{remote_user_id}' should be ALLOWED to invite {self.pro_user_b}"
            assert (
                self.may_invite(remote_user_id, self.pro_user_c, "!madeup:example.com")
                == NOT_SPAM
            ), f"'{remote_user_id}' should be ALLOWED to invite {self.pro_user_c}"

            self.set_permissions_for_user(
                self.pro_user_d,
                PermissionConfig(
                    defaultSetting=PermissionDefaultSetting.ALLOW_ALL,
                    groupExceptions=[{"groupName": GroupName.isInsuredPerson.value}],
                ),
            )
            assert (
                self.may_invite(remote_user_id, self.pro_user_d, "!madeup:example.com")
                == errors.Codes.FORBIDDEN
            ), f"'{remote_user_id}' should be FORBIDDEN to invite {self.pro_user_d} without contact details"


class RemoteEpaModeInviteTest(FederatingModuleApiTestCase):
    """
    These Epa server tests are for invites that happen after the room creation process
    has completed

    Note that if the local server is in 'epa' mode, it means the server 'isInsurance'.
    Therefore, it is the responsibility of the remote server to deny *our* invites.
    Likewise, it is our responsibility to deny *theirs* if they are also 'isInsurance'.

    The second behavior is what we test here
    """

    server_name_for_this_server = INSURANCE_DOMAIN_IN_LIST_FOR_LOCAL

    def prepare(self, reactor: MemoryReactor, clock: Clock, homeserver: HomeServer):
        super().prepare(reactor, clock, homeserver)
        self.epa_user_d = self.register_user("d", "password")
        self.non_existent_user = f"@notarealuser:{self.server_name_for_this_server}"
        self.login("d", "password")

    def default_config(self) -> dict[str, Any]:
        conf = super().default_config()
        assert "modules" in conf, "modules missing from config dict during construction"

        # There should only be a single item in the 'modules' list, since this tests that module
        assert len(conf["modules"]) == 1, "more than one module found in config"

        conf["modules"][0].setdefault("config", {}).update({"tim-type": "epa"})
        return conf

    def may_invite(self, inviter: str, invitee: str, room_id: str):
        req = defer.ensureDeferred(
            self.hs.get_module_api()._callbacks.spam_checker.user_may_invite(
                inviter, invitee, room_id
            )
        )
        self.wait_on_thread(req)
        ret = self.get_success(req)
        if ret == NOT_SPAM:
            return NOT_SPAM
        return ret[0]  # return first code instead of all of them to make assert easier

    def test_invite_from_remote_not_on_fed_list(self) -> None:
        """Tests that an invite from a remote server not in the federation list gets denied"""
        # Add in permissions for one of them, it doesn't work anyway
        self.add_permission_to_a_user(f"@example:not-{DOMAIN_IN_LIST}", self.epa_user_d)

        for remote_user_id in (
            f"@example:not-{DOMAIN_IN_LIST}",
            f"@example2:not-{DOMAIN_IN_LIST}",
        ):
            assert (
                self.may_invite(remote_user_id, self.epa_user_d, "!madeup:example.com")
                == errors.Codes.FORBIDDEN
            ), f"'{remote_user_id}' should be FORBIDDEN to invite {self.epa_user_d}"

    def test_invite_from_remote_users(self) -> None:
        """
        Tests that an invite from a remote server gets accepted when in the federation
        list.
        """
        for remote_user_id in (
            f"@mxid:{DOMAIN_IN_LIST}",
            f"@matrixuri:{DOMAIN_IN_LIST}",
        ):
            local_user = self.epa_user_d

            self.set_permissions_for_user(local_user, PermissionConfig())

            # default
            assert (
                self.may_invite(remote_user_id, local_user, "!madeup:example.com")
                == NOT_SPAM
            ), f"'{remote_user_id}' should be ALLOWED to invite {local_user}"

            # Explicit allow all
            self.set_permissions_for_user(
                local_user,
                PermissionConfig(defaultSetting=PermissionDefaultSetting.ALLOW_ALL),
            )
            assert (
                self.may_invite(remote_user_id, local_user, "!madeup:example.com")
                == NOT_SPAM
            ), f"'{remote_user_id}' should be ALLOWED to invite {local_user} (explicit all)"

            # Explicit block all
            self.set_permissions_for_user(
                local_user,
                PermissionConfig(defaultSetting=PermissionDefaultSetting.BLOCK_ALL),
            )
            assert (
                self.may_invite(remote_user_id, local_user, "!madeup:example.com")
                == errors.Codes.FORBIDDEN
            ), f"'{remote_user_id}' should be BLOCKED from inviting {local_user} (explicit all)"

            # Explicit allow single
            self.set_permissions_for_user(
                local_user,
                PermissionConfig(
                    defaultSetting=PermissionDefaultSetting.BLOCK_ALL,
                    userExceptions={remote_user_id: {}},
                ),
            )
            assert (
                self.may_invite(remote_user_id, local_user, "!madeup:example.com")
                == NOT_SPAM
            ), f"'{remote_user_id}' should be ALLOWED to invite {local_user} (explicit single)"

            # Explicit block single
            self.set_permissions_for_user(
                local_user,
                PermissionConfig(
                    defaultSetting=PermissionDefaultSetting.ALLOW_ALL,
                    userExceptions={remote_user_id: {}},
                ),
            )
            assert (
                self.may_invite(remote_user_id, local_user, "!madeup:example.com")
                == errors.Codes.FORBIDDEN
            ), f"'{remote_user_id}' should be BLOCKED from inviting {local_user} (explicit single)"

            # Explicit allow server
            self.set_permissions_for_user(
                local_user,
                PermissionConfig(
                    defaultSetting=PermissionDefaultSetting.BLOCK_ALL,
                    serverExceptions={UserID.from_string(remote_user_id).domain: {}},
                ),
            )
            assert (
                self.may_invite(remote_user_id, local_user, "!madeup:example.com")
                == NOT_SPAM
            ), f"'{remote_user_id}' should be ALLOWED to invite {local_user} (explicit server)"

            # Explicit block server
            self.set_permissions_for_user(
                local_user,
                PermissionConfig(
                    defaultSetting=PermissionDefaultSetting.ALLOW_ALL,
                    serverExceptions={UserID.from_string(remote_user_id).domain: {}},
                ),
            )
            assert (
                self.may_invite(remote_user_id, local_user, "!madeup:example.com")
                == errors.Codes.FORBIDDEN
            ), f"'{remote_user_id}' should be BLOCKED from inviting {local_user} (explicit server)"

    def test_remote_invite_from_an_insured_domain_fails(self) -> None:
        """
        Test that invites from another insurance domain are rejected with or without
        contact permissions
        """
        for remote_user_id in (
            f"@unknown:{INSURANCE_DOMAIN_IN_LIST}",
            f"@rando-32-b52:{INSURANCE_DOMAIN_IN_LIST}",
        ):
            assert (
                self.may_invite(remote_user_id, self.epa_user_d, "!madeup:example.com")
                == errors.Codes.FORBIDDEN
            ), f"'{remote_user_id}' should be FORBIDDEN to invite {self.epa_user_d}"

            # Add in permissions
            self.add_permission_to_a_user(remote_user_id, self.epa_user_d)

            # ...and try again
            assert (
                self.may_invite(remote_user_id, self.epa_user_d, "!madeup:example.com")
                == errors.Codes.FORBIDDEN
            ), f"'{remote_user_id}' should be FORBIDDEN to invite {self.epa_user_d}"

    def test_invite_to_nonexistent_local_user(self) -> None:
        """Tests that an invite to a local user that does not exist gets denied"""
        for remote_user_id in [
            f"@example:{DOMAIN_IN_LIST}",
            f"@example:{DOMAIN3_IN_LIST}",
            f"@mxid404:{DOMAIN_IN_LIST}",
        ]:
            assert (
                self.may_invite(
                    remote_user_id, self.non_existent_user, "!madeup:example.com"
                )
                == errors.Codes.FORBIDDEN
            ), f"'{remote_user_id}' should be FORBIDDEN to invite {self.non_existent_user}(before permission)"

            # Adding permissions will not help, as the user doesn't exist
            with pytest.raises(
                ValueError,
                match=re.escape(
                    "User @notarealuser:ti-messengertest.dev.ccs.gematik.solutions does not exist on this server."
                ),
            ):
                # Raises with this error:
                # `ValueError: User @notarealuser:tim.test.gematik.de does not exist on this server.`
                # Weirdly, it's not the retrieval of the account data that triggers it,
                # but is the trying to put new account data
                self.add_permission_to_a_user(remote_user_id, self.non_existent_user)

            # May as well give it another go anyway, just to make sure getting the
            # non-existent account data didn't cause it to suddenly work
            assert (
                self.may_invite(
                    remote_user_id, self.non_existent_user, "!madeup:example.com"
                )
                == errors.Codes.FORBIDDEN
            ), f"'{remote_user_id}' should be FORBIDDEN to invite {self.non_existent_user}(after permission)"



# Copyright (C) 2020, 2024 Famedly
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
import re
from typing import Any, ClassVar
from unittest import TestCase

import pytest
from synapse.config import ConfigError

from synapse_invite_checker import InviteChecker
from synapse_invite_checker.types import DefaultPermissionConfig, TimType


class ConfigParsingTestCase(TestCase):
    """
    Test that parsing the config can generate Exceptions.

    We start with a basic configuration, and copy it for each test. That is then
    modified to generate the Exceptions.

    An empty string was chosen to represent a 'falsy' value, but removing the
    key has the same effect. May use either interchangeably
    """

    config: ClassVar[dict[str, Any]] = {
        "tim-type": "pro",
        "federation_list_url": "https://localhost:8080",
        "federation_list_client_cert": "tests/certs/client.pem",
        "gematik_ca_baseurl": "https://download-ref.tsl.ti-dienste.de/",
        "allowed_room_versions": ["9", "10"],
    }

    def test_tim_type_is_not_case_sensitive(self) -> None:
        test_config = self.config.copy()
        test_config.update({"tim-type": "ePA"})
        assert InviteChecker.parse_config(test_config).tim_type == TimType.EPA

    def test_tim_type_defaults_to_pro_mode(self) -> None:
        test_config = self.config.copy()
        test_config.pop("tim-type")
        assert InviteChecker.parse_config(test_config).tim_type == TimType.PRO

    def test_incorrect_tim_type_raises(self) -> None:
        test_config = self.config.copy()
        test_config.update({"tim-type": "fake"})
        with pytest.raises(ConfigError):
            InviteChecker.parse_config(test_config)

    def test_missing_fed_list_or_gematik_ca_url_raises(self) -> None:
        test_config = self.config.copy()
        test_config.update({"federation_list_url": ""})
        with pytest.raises(Exception, match="Incomplete federation list config"):
            InviteChecker.parse_config(test_config)

        test_config = self.config.copy()
        test_config.update({"gematik_ca_baseurl": ""})
        with pytest.raises(Exception, match="Incomplete federation list config"):
            InviteChecker.parse_config(test_config)

    def test_missing_fed_list_client_certs_raises(self) -> None:
        """
        Test that missing client certs for the Federation List raises an exception
        when using HTTPS with default mTLS settings (required)
        """
        test_config = self.config.copy()
        test_config.update({"federation_list_client_cert": ""})
        with pytest.raises(
            Exception,
            match=re.escape(
                "Federation list config requires an mtls (PEM) cert for https connections when mTLS is required"
            ),
        ):
            InviteChecker.parse_config(test_config)

    def test_missing_fed_list_client_certs_with_mtls_disabled(self) -> None:
        """
        Test that missing client certs for the Federation List is accepted
        when using HTTPS but mTLS is disabled
        """
        test_config = self.config.copy()
        test_config.update(
            {
                "federation_list_client_cert": "",
                "federation_list_require_mtls": False,
            }
        )
        # Should not raise an exception
        assert InviteChecker.parse_config(test_config)

    def test_federation_list_require_mtls_defaults_to_true(self) -> None:
        """
        Test that federation_list_require_mtls defaults to True for backward compatibility
        """
        test_config = self.config.copy()
        config = InviteChecker.parse_config(test_config)
        assert (
            config.federation_list_require_mtls
        ), "federation_list_require_mtls should default to True"

    def test_missing_fed_list_client_certs_is_accepted_if_fed_scheme_is_http(
        self,
    ) -> None:
        """
        'federation_list_url' must be 'http' if Federation List Client certs are missing.
        """
        test_config = self.config.copy()
        test_config.update(
            {
                "federation_list_client_cert": "",
                "federation_list_url": "http://localhost:8080",
            }
        )

        assert InviteChecker.parse_config(test_config), "Exception maybe?"

    def test_allowed_room_versions_is_not_a_list(self) -> None:
        test_config = self.config.copy()

        assert InviteChecker.parse_config(test_config)

        # Nope, not a list
        test_config.update({"allowed_room_versions": "9"})
        with pytest.raises(ConfigError):
            InviteChecker.parse_config(test_config)

        # Nope, not a list
        test_config.update({"allowed_room_versions": "{9}"})
        with pytest.raises(ConfigError):
            InviteChecker.parse_config(test_config)

        # Nope, not a recognized list
        test_config.update({"allowed_room_versions": "9, 10"})
        with pytest.raises(ConfigError):
            InviteChecker.parse_config(test_config)

        # Nope, not a list
        test_config.update({"allowed_room_versions": "9 10"})
        with pytest.raises(ConfigError):
            InviteChecker.parse_config(test_config)

        # This one is okay, these are integers and can be coerced into strings
        test_config.update({"allowed_room_versions": [9, 10]})
        assert InviteChecker.parse_config(test_config)

        # This is allowed
        test_config.update({"allowed_room_versions": ["8"]})
        assert InviteChecker.parse_config(test_config)

    def test_parse_duration_parameters(self) -> None:
        test_config = self.config.copy()
        test_config.update({"room_scan_run_interval": "bad value"})
        # After Synapse release 1.124.0, this became a TypeError
        with pytest.raises((ValueError, TypeError)):
            InviteChecker.parse_config(test_config)

        test_config = self.config.copy()
        # Specifically use a word that has a final letter that matches one recognized
        # by parse_duration()
        test_config.update({"room_scan_run_interval": "why"})
        with pytest.raises(
            ValueError, match=re.escape("invalid literal for int() with base 10: 'wh'")
        ):
            InviteChecker.parse_config(test_config)

    def test_dict_unexpectedly_is_something_else_raises(self) -> None:
        test_config = self.config.copy()
        # Shouldn't work if set to a string
        test_config.update({"insured_only_room_scan": "bad value"})
        with pytest.raises(ConfigError):
            InviteChecker.parse_config(test_config)

        # Shouldn't work if set to a list of strings
        test_config.update({"insured_only_room_scan": ["what", "is", "a", "list?"]})
        with pytest.raises(ConfigError):
            InviteChecker.parse_config(test_config)

        test_config = self.config.copy()
        # Shouldn't work if set to a string
        test_config.update({"inactive_room_scan": "not a dict"})
        with pytest.raises(ConfigError):
            InviteChecker.parse_config(test_config)

        # Shouldn't work if set to a list of strings
        test_config.update(
            {"inactive_room_scan": ["lists", "are", "only", "good", "on", "mondays"]}
        )
        with pytest.raises(ConfigError):
            InviteChecker.parse_config(test_config)

    def test_public_room_override_defaults_to_true(self) -> None:
        test_config = self.config.copy()
        config = InviteChecker.parse_config(test_config)
        assert (
            config.override_public_room_federation
        ), "`override_public_room_federation` should default to 'True'"

    def test_public_room_override_can_be_disabled(self) -> None:
        test_config = self.config.copy()
        test_config.update({"override_public_room_federation": False})
        config = InviteChecker.parse_config(test_config)
        assert (
            config.override_public_room_federation is False
        ), "`override_public_room_federation` should be `False`'"

    def test_public_room_override_raises(self) -> None:
        test_config = self.config.copy()
        # Although a boolean is an int, an int is not a boolean
        test_config.update({"override_public_room_federation": "0"})
        with pytest.raises(ConfigError):
            InviteChecker.parse_config(test_config)

        # No silly strings. Shame, really....
        test_config = self.config.copy()
        test_config.update({"override_public_room_federation": "nope"})
        with pytest.raises(ConfigError):
            InviteChecker.parse_config(test_config)

    def test_prohibit_world_readable_rooms_override_defaults_to_true(self) -> None:
        test_config = self.config.copy()
        config = InviteChecker.parse_config(test_config)
        assert (
            config.prohibit_world_readable_rooms
        ), "`prohibit_world_readable_rooms` should default to 'True'"

    def test_prohibit_world_readable_rooms_override_can_be_disabled(self) -> None:
        test_config = self.config.copy()
        test_config.update({"prohibit_world_readable_rooms": False})
        config = InviteChecker.parse_config(test_config)
        assert (
            config.prohibit_world_readable_rooms is False
        ), "`prohibit_world_readable_rooms` should be `False`'"

    def test_prohibit_world_readable_rooms_override_raises(self) -> None:
        test_config = self.config.copy()
        # Although a boolean is an int, an int is not a boolean
        test_config.update({"prohibit_world_readable_rooms": "1"})
        with pytest.raises(ConfigError):
            InviteChecker.parse_config(test_config)

        # No silly strings. Shame, really....
        test_config = self.config.copy()
        test_config.update({"prohibit_world_readable_rooms": "hi_mom!"})
        with pytest.raises(ConfigError):
            InviteChecker.parse_config(test_config)

    def test_federation_list_require_mtls_raises_for_non_boolean(self) -> None:
        """Verify that non-boolean values for federation_list_require_mtls raise ConfigError"""
        test_config = self.config.copy()
        # String that's not a boolean
        test_config.update({"federation_list_require_mtls": "maybe"})
        with pytest.raises(ConfigError):
            InviteChecker.parse_config(test_config)

        # Integer is not a boolean
        test_config = self.config.copy()
        test_config.update({"federation_list_require_mtls": 1})
        with pytest.raises(ConfigError):
            InviteChecker.parse_config(test_config)

    def test_invalid_url_scheme_raises(self) -> None:
        """
        Test that using an invalid URL scheme (neither http nor https) raises an exception
        when creating an MtlsPolicy
        """
        from synapse_invite_checker.config import InviteCheckerConfig
        from synapse_invite_checker.invite_checker import MtlsPolicy

        # Create a config with an invalid URL scheme
        config = InviteCheckerConfig(DefaultPermissionConfig())
        config.federation_list_url = "ftp://localhost:8080"
        with pytest.raises(Exception, match="URL scheme must be either http or https"):
            MtlsPolicy(config)

        config.federation_list_url = "ws://localhost:8080"
        with pytest.raises(Exception, match="URL scheme must be either http or https"):
            MtlsPolicy(config)

        config.federation_list_url = "file:///etc/passwd"
        with pytest.raises(Exception, match="URL scheme must be either http or https"):
            MtlsPolicy(config)

    def test_mtls_policy_creator_for_netloc_bytes_hostname(self) -> None:
        """
        Test that creatorForNetloc correctly handles bytes hostname by encoding url.hostname to UTF-8
        """
        from synapse_invite_checker.config import InviteCheckerConfig
        from synapse_invite_checker.invite_checker import MtlsPolicy

        # Create a config with valid URL
        config = InviteCheckerConfig(DefaultPermissionConfig())
        config.federation_list_url = "https://example.com:8443"
        config.federation_list_client_cert = ""
        config.federation_list_require_mtls = False

        policy = MtlsPolicy(config)

        # Test with matching bytes hostname (that corresponds to the encoded url.hostname) and port
        options = policy.creatorForNetloc(b"example.com", 8443)
        assert options == policy.options

        # Test with non-matching bytes hostname
        with pytest.raises(Exception, match="Invalid connection attempt"):
            policy.creatorForNetloc(b"other.com", 8443)

        # Test with invalid bytes hostname
        invalid_bytes = b"\xff\xfe\xfd\xfc"  # Not valid UTF-8
        with pytest.raises(Exception, match="Invalid connection attempt"):
            policy.creatorForNetloc(invalid_bytes, 8443)

    def test_unexpected_non_booleans(self) -> None:
        test_config = self.config.copy()
        # Although a boolean is an int, an int is not a boolean
        test_config.update({"limit_reactions": "1"})
        with pytest.raises(ConfigError):
            InviteChecker.parse_config(test_config)

        test_config = self.config.copy()
        # Although a boolean is an int, an int is not a boolean
        test_config.update({"limit_reactions": 1})
        with pytest.raises(ConfigError):
            InviteChecker.parse_config(test_config)
        test_config = self.config.copy()

        # Although a boolean is an int, an int is not a boolean
        test_config.update({"limit_reactions": "hi mom!"})
        with pytest.raises(ConfigError):
            InviteChecker.parse_config(test_config)



# Copyright (C) 2025 Famedly
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
from http import HTTPStatus

from parameterized import parameterized
from synapse.server import HomeServer
from synapse.util.clock import Clock
from twisted.internet.testing import MemoryReactor

from tests.base import FederatingModuleApiTestCase, construct_extra_content
from tests.server import make_request


class RoomVersionCreateRoomTest(FederatingModuleApiTestCase):
    """
    Tests for limiting room versions when creating rooms. Use the defaults of room
    versions "9" or "10"
    """

    def prepare(self, reactor: MemoryReactor, clock: Clock, homeserver: HomeServer):
        super().prepare(reactor, clock, homeserver)

        self.user_a = self.register_user("a", "password")
        self.access_token_a = self.login("a", "password")

        self.admin_b = self.register_user("b", "password", admin=True)
        self.access_token_b = self.login("b", "password")

    def user_create_room(
        self,
        is_public: bool = False,
        room_ver: str | None = None,
        expect_code: int = HTTPStatus.OK,
    ) -> str | None:
        """
        Helper to send an api request with a full set of required additional room state
        to the room creation matrix endpoint. Returns a room_id if successful.

        This differs from the create_local_room() helper, as it allows to designate
        the room version and the expected response code. It also will subvert the
        invitee_list, as it is not useful when determining if the room would be created.
        """
        # Hide the assertion from create_room_as() when the error code is unexpected. It
        # makes errors for the tests less clear when all we get is the http response
        return self.helper.create_room_as(
            self.user_a,
            is_public=is_public,
            room_version=room_ver,
            tok=self.access_token_a,
            expect_code=expect_code,
            extra_content=construct_extra_content(self.user_a, []),
        )

    def upgrade_room_to_version(
        self,
        _room_id: str,
        room_version: str,
        tok: str | None = None,
    ) -> str | None:
        """
        Upgrade a room.

        Args:
            _room_id
            room_version: The room version to upgrade the room to.
            tok: The access token to use in the request.
        Returns:
            The ID of the newly created room, or None if the request failed.
        """
        path = f"/_matrix/client/r0/rooms/{_room_id}/upgrade"
        content = {"new_version": room_version}

        channel = make_request(
            self.reactor,
            self.site,
            "POST",
            path,
            content,
            access_token=tok,
        )

        return channel.json_body.get("replacement_room")

    @parameterized.expand([("public", True), ("private", False)])
    def test_create_room_fails(self, _label: str, is_public: bool) -> None:
        """
        Test that most generic ways of not doing a room version string, and a room
        version that is outside of what is wanted, fail
        """
        self.user_create_room(
            is_public=is_public,
            room_ver="8",
            expect_code=HTTPStatus.BAD_REQUEST,
        )

        self.user_create_room(
            is_public=is_public,
            room_ver=8,  # type: ignore[arg-type]
            expect_code=HTTPStatus.BAD_REQUEST,
        )

        self.user_create_room(
            is_public=is_public,
            room_ver="11",
            expect_code=HTTPStatus.BAD_REQUEST,
        )

        self.user_create_room(
            is_public=is_public,
            room_ver=11,  # type: ignore[arg-type]
            expect_code=HTTPStatus.BAD_REQUEST,
        )

        self.user_create_room(
            is_public=is_public,
            room_ver="bad_version",
            expect_code=HTTPStatus.BAD_REQUEST,
        )

    @parameterized.expand([("public", True), ("private", False)])
    def test_create_room_succeeds(self, _label: str, is_public: bool) -> None:
        """
        Tests that a room version that is allowed succeeds
        """
        self.user_create_room(
            is_public=is_public,
            room_ver="9",
            expect_code=HTTPStatus.OK,
        )
        self.user_create_room(
            is_public=is_public,
            room_ver="10",
            expect_code=HTTPStatus.OK,
        )

    @parameterized.expand([("public", True), ("private", False)])
    def test_room_upgrades(self, _label: str, is_public: bool) -> None:
        """
        Test room upgrades fail outside of defaults
        """
        # 9 -> 9 works
        room_id = self.user_create_room(is_public=is_public, room_ver="9")
        assert room_id
        room_id = self.upgrade_room_to_version(room_id, "9", self.access_token_a)
        assert room_id

        # 10 -> 10 works
        room_id = self.user_create_room(is_public=is_public, room_ver="10")
        assert room_id
        room_id = self.upgrade_room_to_version(room_id, "10", self.access_token_a)
        assert room_id

        # 9 -> 10 works
        room_id = self.user_create_room(is_public=is_public, room_ver="9")
        assert room_id
        room_id = self.upgrade_room_to_version(room_id, "10", self.access_token_a)
        assert room_id

        # 9 -> 8 doesn't work
        room_id = self.user_create_room(is_public=is_public, room_ver="9")
        assert room_id

        room_id = self.upgrade_room_to_version(room_id, "8", self.access_token_a)
        assert room_id is None

        # 9 -> 8 requires an admin
        room_id = self.helper.create_room_as(
            self.admin_b,
            is_public=is_public,
            room_version="9",
            tok=self.access_token_b,
            expect_code=200,
            extra_content=construct_extra_content(self.admin_b, []),
        )
        assert room_id

        room_id = self.upgrade_room_to_version(room_id, "8", self.access_token_b)
        assert room_id



# Copyright 2019-2021 The Matrix.org Foundation C.I.C.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Utilities for running the unit tests
"""

import json
import sys
import warnings
from binascii import unhexlify
from collections.abc import Awaitable, Callable
from typing import TYPE_CHECKING, Protocol, TypeVar

import attr
from synapse.types import JsonSerializable
from twisted.internet.interfaces import IProtocol
from twisted.python.failure import Failure
from twisted.web.client import ResponseDone
from twisted.web.http import RESPONSES
from twisted.web.http_headers import Headers
from twisted.web.iweb import IResponse
from zope.interface import implementer

if TYPE_CHECKING:
    from sys import UnraisableHookArgs

TV = TypeVar("TV")

# Only use this server name for our PRO test instance
SERVER_NAME_FROM_LIST = "tim.test.gematik.de"
# Use a domain found in our raw fedlist data instead of inserting a fake one
DOMAIN_IN_LIST = "timo.staging.famedly.de"
DOMAIN2_IN_LIST = "tim-alpha.staging.famedly.de"
DOMAIN3_IN_LIST = "messenger.spilikin.dev"
# Only use this server name for our EPA test instance
INSURANCE_DOMAIN_IN_LIST = "cirosec.de"
# And another 'isInsurance' domain for testing local
INSURANCE_DOMAIN_IN_LIST_FOR_LOCAL = "ti-messengertest.dev.ccs.gematik.solutions"


def get_awaitable_result(awaitable: Awaitable[TV]) -> TV:
    """Get the result from an Awaitable which should have completed

    Asserts that the given awaitable has a result ready, and returns its value
    """
    i = awaitable.__await__()
    try:
        next(i)
    except StopIteration as e:
        # awaitable returned a result
        return e.value

    # if next didn't raise, the awaitable hasn't completed.
    msg = "awaitable has not yet completed"
    raise Exception(msg)


def setup_awaitable_errors() -> Callable[[], None]:
    """
    Convert warnings from a non-awaited coroutines into errors.
    """
    warnings.simplefilter("error", RuntimeWarning)

    # State shared between unraisablehook and check_for_unraisable_exceptions.
    unraisable_exceptions = []
    orig_unraisablehook = sys.unraisablehook

    def unraisablehook(unraisable: "UnraisableHookArgs") -> None:
        unraisable_exceptions.append(unraisable.exc_value)

    def cleanup() -> None:
        """
        A method to be used as a clean-up that fails a test-case if there are any new unraisable exceptions.
        """
        sys.unraisablehook = orig_unraisablehook
        if unraisable_exceptions:
            exc = unraisable_exceptions.pop()
            assert exc is not None
            raise exc

    sys.unraisablehook = unraisablehook

    return cleanup


class IResponseProto(Protocol):
    def deliverBody(self, protocol: IProtocol) -> None: ...


# Type ignore: it does not fully implement IResponse, but is good enough for tests
@implementer(IResponse)
@attr.s(slots=True, frozen=True, auto_attribs=True)
class FakeResponse:  # type: ignore[misc]
    """A fake twisted.web.IResponse object

    there is a similar class at treq.test.test_response, but it lacks a `phrase`
    attribute, and didn't support deliverBody until recently.
    """

    version: tuple[bytes, int, int] = (b"HTTP", 1, 1)

    # HTTP response code
    code: int = 200

    # body of the response
    body: bytes = b""

    headers: Headers = attr.Factory(Headers)

    @property
    def phrase(self) -> bytes:
        return RESPONSES.get(self.code, b"Unknown Status")

    @property
    def length(self) -> int:
        return len(self.body)

    def deliverBody(self, protocol: IProtocol) -> None:
        protocol.dataReceived(self.body)
        protocol.connectionLost(Failure(ResponseDone()))

    @classmethod
    def json(cls, *, code: int = 200, payload: JsonSerializable) -> "FakeResponse":
        headers = Headers({"Content-Type": ["application/json"]})
        body = json.dumps(payload).encode("utf-8")
        return cls(code=code, body=body, headers=headers)  # type: ignore[call-arg]


# A small image used in some tests.
#
# Resolution: 1x1, MIME type: image/png, Extension: png, Size: 67 B
SMALL_PNG = unhexlify(
    b"89504e470d0a1a0a0000000d4948445200000001000000010806"
    b"0000001f15c4890000000a49444154789c63000100000500010d"
    b"0a2db40000000049454e44ae426082"
)



# Copyright (C) 2020, 2024 Famedly
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
from http import HTTPStatus

from synapse.server import HomeServer
from synapse.util.clock import Clock
from twisted.internet.testing import MemoryReactor

import tests.unittest as synapse_test
from tests.base import FederatingModuleApiTestCase


class MessengerInfoTestCase(FederatingModuleApiTestCase):
    def prepare(self, reactor: MemoryReactor, clock: Clock, homeserver: HomeServer):
        super().prepare(reactor, clock, homeserver)
        self.user_a = self.register_user("a", "password")
        self.access_token = self.login("a", "password")

    def test_no_access_token_is_denied(self) -> None:
        channel = self.make_request(
            method="GET",
            path="/tim-information",
            shorthand=False,
        )

        assert (
            channel.code == HTTPStatus.UNAUTHORIZED
        ), "Request should fail with no access token"

    def test_default_operator_contact_info_resource(self) -> None:
        """Tests that the messenger operator contact info resource is accessible"""

        channel = self.make_request(
            method="GET",
            path="/tim-information",
            access_token=self.access_token,
            shorthand=False,
        )

        assert channel.code == HTTPStatus.OK, channel.result
        assert channel.json_body["title"] == "Invite Checker module by Famedly"
        assert channel.json_body["description"] == "Invite Checker module by Famedly"
        assert channel.json_body["contact"] == "info@famedly.com"
        assert channel.json_body["version"], "Version returned"

    @synapse_test.override_config(
        {
            "modules": [
                {
                    "module": "synapse_invite_checker.InviteChecker",
                    "config": {
                        "title": "abc",
                        "description": "def",
                        "contact": "ghi",
                        "federation_list_url": "https://localhost:8080",
                        "federation_list_client_cert": "tests/certs/client.pem",
                        "gematik_ca_baseurl": "https://download-ref.tsl.ti-dienste.de/",
                        "allowed_room_versions": ["9", "10"],
                    },
                }
            ]
        }
    )
    def test_custom_operator_contact_info_resource(self) -> None:
        """Tests that the registered info resource is accessible and has the configured values"""

        channel = self.make_request(
            method="GET",
            path="/tim-information",
            access_token=self.access_token,
            shorthand=False,
        )

        assert channel.code == HTTPStatus.OK, channel.result
        assert channel.json_body["title"] == "abc"
        assert channel.json_body["description"] == "def"
        assert channel.json_body["contact"] == "ghi"
        assert channel.json_body["version"], "Version returned"


class MessengerIsInsuranceResourceTest(FederatingModuleApiTestCase):
    def prepare(self, reactor: MemoryReactor, clock: Clock, homeserver: HomeServer):
        super().prepare(reactor, clock, homeserver)
        self.user_a = self.register_user("a", "password")
        self.access_token = self.login("a", "password")

    def test_pro_isInsurance_returns_expected(self) -> None:
        """Tests that Pro mode returns expected response"""
        channel = self.make_request(
            method="GET",
            path="/tim-information/v1/server/isInsurance?serverName=cirosec.de",
            shorthand=False,
        )

        assert (
            channel.code == HTTPStatus.UNAUTHORIZED
        ), "Request should fail with no access token"

        channel = self.make_request(
            method="GET",
            path="/tim-information/v1/server/isInsurance",
            access_token=self.access_token,
            shorthand=False,
        )

        assert (
            channel.code == HTTPStatus.BAD_REQUEST
        ), "Request should have a parameter missing"

        channel = self.make_request(
            method="GET",
            path="/tim-information/v1/server/isInsurance?serverName=cirosec.de",
            access_token=self.access_token,
            shorthand=False,
        )

        assert channel.code == HTTPStatus.OK, channel.result
        assert channel.json_body[
            "isInsurance"
        ], "isInsurance is FALSE when it should be TRUE"

        channel = self.make_request(
            method="GET",
            path="/tim-information/v1/server/isInsurance?serverName=timo.staging.famedly.de",
            access_token=self.access_token,
            shorthand=False,
        )

        assert channel.code == HTTPStatus.OK, channel.result
        assert not channel.json_body[
            "isInsurance"
        ], "isInsurance is TRUE when it should be FALSE"

    @synapse_test.override_config(
        {
            "modules": [
                {
                    "module": "synapse_invite_checker.InviteChecker",
                    "config": {
                        "tim-type": "epa",
                        "title": "abc",
                        "description": "def",
                        "contact": "ghi",
                        "federation_list_url": "https://localhost:8080",
                        "federation_list_client_cert": "tests/certs/client.pem",
                        "gematik_ca_baseurl": "https://download-ref.tsl.ti-dienste.de/",
                        "allowed_room_versions": ["9", "10"],
                    },
                }
            ]
        }
    )
    def test_epa_isInsurance_returns_expected(self) -> None:
        """Tests that ePA mode returns expected response"""

        channel = self.make_request(
            method="GET",
            path="/tim-information/v1/server/isInsurance?serverName=cirosec.de",
            shorthand=False,
        )

        assert channel.code == HTTPStatus.NOT_FOUND, "Endpoint shouldn't exist"

        channel = self.make_request(
            method="GET",
            path="/tim-information/v1/server/isInsurance?serverName=timo.staging.famedly.de",
            shorthand=False,
        )

        assert channel.code == HTTPStatus.NOT_FOUND, "Endpoint shouldn't exist"


class MessengerFindByIkResourceTestCase(FederatingModuleApiTestCase):
    def prepare(self, reactor: MemoryReactor, clock: Clock, homeserver: HomeServer):
        super().prepare(reactor, clock, homeserver)
        self.user_a = self.register_user("a", "password")
        self.access_token = self.login("a", "password")

    def test_no_auth(self) -> None:
        channel = self.make_request(
            method="GET",
            path="/tim-information/v1/server/findByIk?ikNumber=1",
            shorthand=False,
        )

        assert (
            channel.code == HTTPStatus.UNAUTHORIZED
        ), "Request should fail with no access token"

    def test_not_found(self) -> None:
        channel = self.make_request(
            method="GET",
            path="/tim-information/v1/server/findByIk?ikNumber=1",
            access_token=self.access_token,
            shorthand=False,
        )

        assert channel.code == HTTPStatus.NOT_FOUND, channel.result

    def test_no_parameter(self) -> None:
        channel = self.make_request(
            method="GET",
            path="/tim-information/v1/server/findByIk",
            access_token=self.access_token,
            shorthand=False,
        )

        assert channel.code == HTTPStatus.BAD_REQUEST, channel.result

    @synapse_test.override_config(
        {
            "modules": [
                {
                    "module": "synapse_invite_checker.InviteChecker",
                    "config": {
                        "tim-type": "epa",
                        "title": "abc",
                        "description": "def",
                        "contact": "ghi",
                        "federation_list_url": "https://localhost:8080",
                        "federation_list_client_cert": "tests/certs/client.pem",
                        "gematik_ca_baseurl": "https://download-ref.tsl.ti-dienste.de/",
                        "allowed_room_versions": ["9", "10"],
                    },
                }
            ]
        }
    )
    def test_unavailable_for_epa(self) -> None:
        # In EPA mode, it shouldn't matter what we do, should always be a 404
        channel = self.make_request(
            method="GET",
            path="/tim-information/v1/server/findByIk?ikNumber=1",
            access_token=self.access_token,
            shorthand=False,
        )

        assert channel.code == HTTPStatus.NOT_FOUND, channel.result



# Copyright (C) 2020,2023 Famedly
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
from dataclasses import dataclass
from enum import Enum, auto
from functools import cached_property
from typing import Any, Final

from pydantic import BaseModel, ConfigDict, Field, computed_field, field_validator
from synapse.types import UserID


class PermissionDefaultSetting(Enum):
    ALLOW_ALL = "allow all"
    BLOCK_ALL = "block all"


class GroupName(Enum):
    isInsuredPerson = "isInsuredPerson"


class PermissionConfig(BaseModel):
    model_config = ConfigDict(
        extra="ignore",
        allow_inf_nan=False,
        use_enum_values=True,
    )

    # This is where we will set the default permission setting now. This only is used
    # when the template is not produced in the configuration, per README.md
    defaultSetting: PermissionDefaultSetting = PermissionDefaultSetting.ALLOW_ALL
    # If either of these two exist, they should contain a dict with the key as the
    # exception and then an empty dict inside "for future expansion"
    serverExceptions: dict[str, dict] = Field(default_factory=dict)
    # If there is a key inside userExceptions, it needs to be sure to start with a '@'.
    # Should we validate this or trust the client app does the right thing?
    userExceptions: dict[str, dict] = Field(default_factory=dict)
    # This is slightly different to the two just above, as it uses a list to contain the
    # dict. At time of writing there is exactly one option, and it is part of the
    # GroupName enum class just above
    groupExceptions: list[dict[str, str]] = Field(default_factory=list)

    def dump(self) -> dict[str, Any]:
        # exclude_unset=True strips out the attributes that were never set, and thus
        # will not appear in the JSON. It does not touch empty dict sub-attributes only
        # top level!!
        # mode="json" which effectively keeps the camel casing for us(as that is a
        # gematik requirement) and makes sure the keys remain as strings.
        return self.model_dump(mode="json", exclude_unset=True)

    def is_allow_all(self):
        return self.defaultSetting != PermissionDefaultSetting.BLOCK_ALL.value

    def is_mxid_allowed_to_contact(self, mxid: str, is_mxid_epa: bool) -> bool:
        """
        The main test for allowing or blocking an invitation.
        """
        mxid_domain = UserID.from_string(mxid).domain
        allowed = self.is_allow_all()

        if (
            is_mxid_epa
            and self.is_group_excepted(GroupName.isInsuredPerson)
            or mxid_domain in self.serverExceptions
            or mxid in self.userExceptions
        ):
            allowed = not allowed

        return allowed

    def is_group_excepted(self, group_name: GroupName) -> bool:
        return any(
            True
            for groupException in self.groupExceptions
            if groupException.get("groupName") == group_name.value
        )


LOCAL_SERVER_TEMPLATE: Final[str] = "@LOCAL_SERVER@"


class DefaultPermissionConfig(PermissionConfig):
    """
    Extend PermissionConfig to include a few additional necessities. Use a
    @field_validator() call to set the objects required per gematik spec on the sub-keys
    of serverExceptions and userExceptions. This way can avoid having to hard-wire a '
    {}' into the generated configuration for the module
    """

    @field_validator("serverExceptions", "userExceptions", mode="before")
    @staticmethod
    def transform_field(keys: dict) -> dict:
        """
        This will run on model_validate(and it's variants) to transform the fields to
        include the required empty object as the value per gematik spec.
        """
        # Use this to transform the fields in PermissionConfig from(for example):
        # {"serverExceptions": {"server_name.com": None}}
        # to
        # {"serverExceptions": {"server_name.com": {}}}
        for key, value in dict(keys).items():
            if value is None:
                keys[key] = {}
        return keys

    def maybe_update_server_exceptions(self, local_server_name: str) -> None:
        """
        Swap out the template variable assigned by LOCAL_SERVER_TEMPLATE(above) to
        register the *actual* local server name
        Args:
            local_server_name: the local server name

        Returns: None
        """
        if LOCAL_SERVER_TEMPLATE in self.serverExceptions:
            self.serverExceptions.setdefault(
                local_server_name, self.serverExceptions.get(LOCAL_SERVER_TEMPLATE, {})
            )
            del self.serverExceptions[LOCAL_SERVER_TEMPLATE]


class FederationDomain(BaseModel):
    model_config = ConfigDict(
        strict=True, frozen=True, extra="ignore", allow_inf_nan=False
    )

    domain: str
    telematikID: str
    # timAnbieter is now not required as part of the gematik spec. The invite checker
    # does not use this data itself.
    timAnbieter: str | None = None
    isInsurance: bool
    # ik gets marked as 'strict=False' as not all domains will have that data
    ik: list[str] = Field(default_factory=list, strict=False)


class FederationList(BaseModel):
    model_config = ConfigDict(
        strict=True, frozen=True, extra="ignore", allow_inf_nan=False
    )

    domainList: list[FederationDomain]

    @computed_field  # type: ignore[prop-decorator]
    @cached_property
    def _domains_on_list(self) -> set[str]:
        """
        The deduplicated domains found on the Federation List
        """
        return {domain_data.domain for domain_data in self.domainList}

    @computed_field  # type: ignore[prop-decorator]
    @cached_property
    def _insurance_domains_on_list(self) -> set[str]:
        """
        Only the domains that are also type 'isInsurance'
        """
        return {
            domain_data.domain
            for domain_data in self.domainList
            if domain_data.isInsurance
        }

    @computed_field  # type: ignore[prop-decorator]
    @cached_property
    def _ik_to_domain(self) -> dict[str, str]:
        """
        A reverse lookup mapping for ik->serverName
        """
        ik_mapping = {}
        for domain in self.domainList:
            if domain.isInsurance:
                # I believe that this creates an empty dict if the data isn't present
                ik_mapping.update({ik: domain.domain for ik in domain.ik})
        return ik_mapping

    def allowed(self, domain: str) -> bool:
        """
        Compare against the domains from the Federation List to determine if they are allowed
        """
        return domain in self._domains_on_list

    def is_insurance(self, domain: str) -> bool:
        """
        Is this domain specifically designated as 'isInsurance'
        """
        return domain in self._insurance_domains_on_list

    def get_domain_from_ik(self, ik: str) -> str | None:
        return self._ik_to_domain.get(ik)


class TimType(Enum):
    PRO = auto()
    EPA = auto()


class PermissionConfigType(Enum):
    EPA_ACCOUNT_DATA_TYPE = "de.gematik.tim.account.permissionconfig.epa.v1"
    PRO_ACCOUNT_DATA_TYPE = "de.gematik.tim.account.permissionconfig.pro.v1"


@dataclass(slots=True)
class EpaRoomTimestampResults:
    """
    Collection of timestamps used to decide if a room should have members removed.
    Used for EPA server rooms. Recall that invite/leave timestamps are *only* seeded
    from pro users membership events

    Attributes:
        last_invite_in_room: The timestamp for the newest 'invite' membership in the
            room. Only used if there was no detected 'leave' event.
        last_leave_in_room: The timestamp of the newest 'leave' membership in the room.
            The preferred timestamp to acquire.
        room_creation_ts: If nothing else, use the room creation as a sentinel. The slim
            possibility exists that no 'leave' event exists and any 'invite' may have
            failed. This is allowed to give a user time to try again before they do not
            have access to this room removed.
    """

    last_invite_in_room: int | None = None
    last_leave_in_room: int | None = None
    room_creation_ts: int | None = None



# Copyright (C) 2025 Famedly
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
import logging
from http import HTTPStatus
from typing import Any

from synapse.server import HomeServer
from synapse.util.clock import Clock
from twisted.internet.testing import MemoryReactor

from tests.base import FederatingModuleApiTestCase
from tests.test_utils import INSURANCE_DOMAIN_IN_LIST_FOR_LOCAL

logger = logging.getLogger(__name__)


class LocalProJoinTestCase(FederatingModuleApiTestCase):
    """
    Tests to verify that we don't break local public/private rooms by accident.
    Specifically, this checks the code for joining a room and not just inviting. This is
    needed for PRO servers as they are allowed to have public rooms. EPA servers do not
    need this test, as they do not allow for local joining.
    """

    # server_name_for_this_server = "tim.test.gematik.de"

    def prepare(self, reactor: MemoryReactor, clock: Clock, homeserver: HomeServer):
        super().prepare(reactor, clock, homeserver)
        self.user_a = self.register_user("a", "password")
        self.user_b = self.register_user("b", "password")
        self.access_token_a = self.login("a", "password")
        self.access_token_b = self.login("b", "password")

    def default_config(self) -> dict[str, Any]:
        conf = super().default_config()
        assert "modules" in conf, "modules missing from config dict during construction"

        # There should only be a single item in the 'modules' list, since this tests that module
        assert len(conf["modules"]) == 1, "more than one module found in config"

        conf["modules"][0].setdefault("config", {}).update({"tim-type": "pro"})
        conf["modules"][0].setdefault("config", {}).update(
            {"default_permissions": {"defaultSetting": "allow all"}}
        )
        return conf

    def test_joining_public_with_invites(self) -> None:
        """Test joining a local public room with invites is allowed"""
        room_id = self.create_local_room(self.user_a, [], is_public=True)
        assert room_id is not None, "Room should have been created"

        self.helper.invite(room_id, self.user_a, self.user_b, tok=self.access_token_a)
        self.helper.join(room_id, self.user_b, tok=self.access_token_b)

    def test_joining_public_no_invites(self) -> None:
        """Test joining a local public room with no invites is allowed"""
        room_id = self.create_local_room(self.user_a, [], is_public=True)
        assert room_id is not None, "Room should have been created"

        self.helper.join(room_id, self.user_b, tok=self.access_token_b)

    def test_rejoining_public_no_invites(self) -> None:
        """Test re-joining a local public room with no invites is allowed"""
        room_id = self.create_local_room(self.user_a, [], is_public=True)
        assert room_id is not None, "Room should have been created"

        self.helper.join(room_id, self.user_b, tok=self.access_token_b)
        self.helper.send(room_id, "Be right back!", tok=self.access_token_b)
        self.helper.leave(room_id, self.user_b, tok=self.access_token_b)
        self.helper.join(room_id, self.user_b, tok=self.access_token_b)
        self.helper.send(room_id, "Sorry about that!", tok=self.access_token_b)

    def test_joining_private_no_invites(self) -> None:
        """Test joining a local private room with no invites is denied"""
        room_id = self.create_local_room(self.user_a, [], is_public=False)
        assert room_id is not None, "Room should have been created"

        self.helper.join(
            room_id,
            self.user_b,
            expect_code=HTTPStatus.FORBIDDEN,
            tok=self.access_token_b,
        )

    def test_joining_private_with_invites(self) -> None:
        """Test joining a local private room with invites is allowed"""
        room_id = self.create_local_room(self.user_a, [], is_public=False)
        assert room_id is not None, "Room should have been created"

        self.helper.invite(room_id, self.user_a, self.user_b, tok=self.access_token_a)
        self.helper.join(room_id, self.user_b, tok=self.access_token_b)

    def test_rejoining_private_with_invites(self) -> None:
        """Test re-joining a local private room with initial invites forbidden"""
        room_id = self.create_local_room(self.user_a, [], is_public=False)
        assert room_id is not None, "Room should have been created"

        self.helper.invite(room_id, self.user_a, self.user_b, tok=self.access_token_a)
        self.helper.join(room_id, self.user_b, tok=self.access_token_b)
        self.helper.send(room_id, "Be right back!", tok=self.access_token_b)
        self.helper.leave(room_id, self.user_b, tok=self.access_token_b)
        # Can not rejoin without an invite
        self.helper.join(
            room_id,
            self.user_b,
            tok=self.access_token_b,
            expect_code=HTTPStatus.FORBIDDEN,
        )


class LocalEpaJoinTestCase(FederatingModuleApiTestCase):
    """
    Tests to verify that we don't break local public/private rooms behavior by accident.
    Specifically, this checks the code for joining a room and not just inviting
    """

    server_name_for_this_server = INSURANCE_DOMAIN_IN_LIST_FOR_LOCAL

    def prepare(self, reactor: MemoryReactor, clock: Clock, homeserver: HomeServer):
        super().prepare(reactor, clock, homeserver)
        self.user_a = self.register_user("a", "password")
        self.user_b = self.register_user("b", "password")
        self.access_token_a = self.login("a", "password")
        self.access_token_b = self.login("b", "password")

    def default_config(self) -> dict[str, Any]:
        conf = super().default_config()
        assert "modules" in conf, "modules missing from config dict during construction"

        # There should only be a single item in the 'modules' list, since this tests that module
        assert len(conf["modules"]) == 1, "more than one module found in config"

        conf["modules"][0].setdefault("config", {}).update({"tim-type": "epa"})
        conf["modules"][0].setdefault("config", {}).update(
            {"default_permissions": {"defaultSetting": "allow all"}}
        )
        return conf

    def test_joining_public_fails(self) -> None:
        """
        Test joining a local public room with or without invites fails

        See comments below for thoughts on why
        """
        room_id = self.create_local_room(self.user_a, [], is_public=True)
        # Public rooms don't exist, so it should be "None" here. Rather hard to invite
        # and join a room without a room ID
        assert room_id is None, "Room should not have been created"

        # Actually you can invite to a room with no ID, but will still fail for us
        # because invites between two EPA members is forbidden
        self.helper.invite(
            room_id,  # type: ignore[arg-type]
            self.user_a,
            self.user_b,
            expect_code=HTTPStatus.FORBIDDEN,
            tok=self.access_token_a,
        )

        # Trying to join a room with a "None" as a room ID returns a 400(because the
        # "room id" does not start with a "!" as it's supposed to). We don't have to
        # test that

    def test_joining_private_no_invites(self) -> None:
        """Test joining a local private room with no invites is denied"""
        room_id = self.create_local_room(self.user_a, [], is_public=False)
        assert room_id is not None, "Room should have been created"

        self.helper.join(
            room_id,
            self.user_b,
            expect_code=HTTPStatus.FORBIDDEN,
            tok=self.access_token_b,
        )

    def test_joining_private_with_invites(self) -> None:
        """Test joining a local private room with invites is denied"""
        room_id = self.create_local_room(self.user_a, [], is_public=False)
        assert room_id is not None, "Room should have been created"

        self.helper.invite(
            room_id,
            self.user_a,
            self.user_b,
            expect_code=HTTPStatus.FORBIDDEN,
            tok=self.access_token_a,
        )
        self.helper.join(
            room_id,
            self.user_b,
            expect_code=HTTPStatus.FORBIDDEN,
            tok=self.access_token_b,
        )



# Copyright (C) 2020,2024 Famedly
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
import re


def invite_checker_pattern(root_prefix: str, path_regex: str):
    path = path_regex.removeprefix("/")
    root = root_prefix.removesuffix("/")
    raw_regex = f"^{root}/{path}"

    # we need to strip the /$, otherwise we can't register for the root of the prefix in a handler...
    if raw_regex.endswith("/$"):
        raw_regex = raw_regex.replace("/$", "$")

    return [re.compile(raw_regex)]



# Copyright (C) 2025 Famedly
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
import logging
from collections.abc import Awaitable, Callable

from pydantic import ValidationError
from synapse.module_api import ModuleApi
from synapse.types import UserID

from synapse_invite_checker.config import InviteCheckerConfig
from synapse_invite_checker.types import PermissionConfig, PermissionConfigType, TimType

logger = logging.getLogger(__name__)


class InviteCheckerPermissionsHandler:
    """
    Used to retrieve and store permissions for users while taking into account the event type
    of configuration as required by gematik spec.

    A great deal of this can be removed after the Contacts REST Api has been removed
    """

    def __init__(
        self,
        api: ModuleApi,
        config: InviteCheckerConfig,
        is_domain_insurance_cb: Callable[[str], Awaitable[bool]],
        default_perms_from_config: PermissionConfig,
    ) -> None:
        self.api = api
        self.config = config
        self.account_data_manager = self.api.account_data_manager
        self.is_domain_insurance = is_domain_insurance_cb
        self.default_perms = default_perms_from_config

    async def get_permissions(self, user_id: str) -> PermissionConfig:
        config_type = self.permission_event_type()
        account_data = await self.account_data_manager.get_global(
            user_id, config_type.value
        )
        if not account_data or not account_data.get("defaultSetting"):
            # Overwrite or set the permissions in three cases(two here, third below):
            # 1. No existing permissions or if they are somehow mis-set as {}
            # 2. The defaultSetting key is missing, indicating a broken permission structure
            permissions = self.default_perms
            await self.update_permissions(user_id, permissions)
        else:
            try:
                permissions = PermissionConfig.model_validate(account_data)
            except ValidationError as e:
                # 3. Somehow the json was incomplete or got mangled, set the default as a
                # 'reset action'
                logger.warning(
                    "Permissions for %s found to be broken, resetting as default: %r",
                    user_id,
                    e,
                )
                permissions = self.default_perms
                await self.update_permissions(user_id, permissions)

        return permissions

    async def update_permissions(
        self,
        user_id: str,
        permissions: PermissionConfig,
    ) -> None:
        config_type = self.permission_event_type()
        await self.account_data_manager.put_global(
            user_id,
            config_type.value,
            permissions.dump(),
        )

    def permission_event_type(self) -> PermissionConfigType:
        """
        Convert the TIM server type into the appropriate user permission event type
        """
        if self.config.tim_type == TimType.EPA:
            config_type = PermissionConfigType.EPA_ACCOUNT_DATA_TYPE
        else:
            # Since we default to PRO, this seems prudent unless a better option
            config_type = PermissionConfigType.PRO_ACCOUNT_DATA_TYPE

        return config_type

    async def is_user_allowed(self, local_user_id: str, remote_mxid: str) -> bool:
        """
        The primary check used by InviteChecker's user_may_invite()
        """
        permissions = await self.get_permissions(local_user_id)
        is_remote_mxid_insurance = await self.is_domain_insurance(
            UserID.from_string(remote_mxid).domain
        )
        return permissions.is_mxid_allowed_to_contact(
            remote_mxid, is_mxid_epa=is_remote_mxid_insurance
        )



# Copyright (C) 2020,2024 Famedly
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
import re
from collections.abc import Awaitable, Callable
from http import HTTPStatus

from synapse.http.servlet import RestServlet, parse_string
from synapse.http.site import SynapseRequest
from synapse.module_api import ModuleApi, errors
from synapse.types import JsonDict

from synapse_invite_checker.config import InviteCheckerConfig
from synapse_invite_checker.rest.base import invite_checker_pattern
from synapse_invite_checker.types import FederationList

# Version of TiMessengerInformation interface. See:
# https://github.com/gematik/api-ti-messenger/blob/main/src/openapi/TiMessengerInformation.yaml
_TMI_schema_version = "1.0.0"

INFO_API_PREFIX = "/tim-information"


def tim_info_patterns(path_regex: str) -> list[re.Pattern]:
    return invite_checker_pattern(INFO_API_PREFIX, path_regex)


class MessengerInfoResource(RestServlet):
    def __init__(self, api: ModuleApi, config: InviteCheckerConfig):
        super().__init__()
        self.api = api
        self.config = config
        self.version = _TMI_schema_version

        self.PATTERNS = tim_info_patterns("/$")

    async def on_GET(self, request: SynapseRequest) -> tuple[int, JsonDict]:
        # This is just to verify the requester is actually authorized
        await self.api.get_user_by_req(request)
        return HTTPStatus.OK, {
            "title": self.config.title,
            "description": self.config.description,
            "contact": self.config.contact,
            "version": self.version,
        }


class MessengerIsInsuranceResource(RestServlet):
    def __init__(
        self,
        api: ModuleApi,
        config: InviteCheckerConfig,
        is_insurance_cb: Callable[[str], Awaitable[bool]],
    ):
        super().__init__()
        self.api = api
        self.config = config
        self.is_insurance_cb = is_insurance_cb

        self.PATTERNS = tim_info_patterns("/v1/server/isInsurance$")

    async def on_GET(self, request: SynapseRequest) -> tuple[int, JsonDict]:
        await self.api.get_user_by_req(request)
        server_name = parse_string(request, "serverName", required=True)
        is_insurance = await self.is_insurance_cb(server_name)
        return HTTPStatus.OK, {
            "isInsurance": is_insurance,
        }


class MessengerFindByIkResource(RestServlet):
    def __init__(
        self,
        api: ModuleApi,
        config: InviteCheckerConfig,
        fed_list_cb: Callable[..., Awaitable[FederationList]],
    ):
        super().__init__()
        self.api = api
        self.config = config
        self.fed_list_cb = fed_list_cb

        self.PATTERNS = tim_info_patterns("/v1/server/findByIk$")

    async def on_GET(self, request: SynapseRequest) -> tuple[int, JsonDict]:
        await self.api.get_user_by_req(request)

        ik_num = parse_string(request, "ikNumber", required=True)
        fed_list = await self.fed_list_cb()
        server_name = fed_list.get_domain_from_ik(ik_num)
        if server_name is None:
            return HTTPStatus.NOT_FOUND, {
                "errcode": errors.Codes.NOT_FOUND,
                "error": "ikNumber was not found",
            }
        return HTTPStatus.OK, {
            "serverName": server_name,
        }
